"""
"""
import numpy as np
from collections import defaultdict
import matplotlib.pyplot as plt

from energy_demand.profiles import load_factors
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import plotting_styles

def plot_cross_graphs(
        base_yr,
        comparison_year,
        regions,
        ed_year_fueltype_regs_yh,
        reg_load_factor_y,
        fueltype_int,
        fueltype_str,
        fig_name,
        label_points,
        plotshow):

    result_dict = defaultdict(dict)

    # -------------------------------------------
    # Get base year modelled demand
    # -------------------------------------------
    for year, fuels in ed_year_fueltype_regs_yh.items():

        if year == base_yr:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['demand_by'][reg_geocode] = np.sum(fuels[fueltype_int][reg_nr])
                result_dict['peak_h_demand_by'][reg_geocode] = np.max(fuels[fueltype_int][reg_nr])

            # Demand across all regs
            result_dict['demand_by_all_regs'] = np.sum(fuels[fueltype_int])

            # Peak demand
            fuel_all_reg_yh = np.sum(fuels[fueltype_int], axis=0)# Sum across all regions
            result_dict['peak_h_demand_by_all_regs'] = np.max(fuel_all_reg_yh)

            # Calculate national load factor
            fuel_all_fueltype_reg_yh = np.sum(fuels, axis=1)# Sum across all regions per fueltype
            load_factor_y = load_factors.calc_lf_y(fuel_all_fueltype_reg_yh)
            result_dict['lf_by_all_regs_av'] = load_factor_y[fueltype_int]

        elif year == comparison_year:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['demand_cy'][reg_geocode] = np.sum(fuels[fueltype_int][reg_nr])
                result_dict['peak_h_demand_cy'][reg_geocode] = np.max(fuels[fueltype_int][reg_nr])

            # Demand across all regs
            result_dict['demand_cy_all_regs'] = np.sum(fuels[fueltype_int])

            # Peak demand
            fuel_all_reg_yh = np.sum(fuels[fueltype_int], axis=0) # Sum across all regions
            result_dict['peak_h_demand_cy_all_regs'] = np.max(fuel_all_reg_yh)

            # Calculate national load factor
            fuel_all_fueltype_reg_yh = np.sum(fuels, axis=1)# Sum across all regions per fueltype
            load_factor_y = load_factors.calc_lf_y(fuel_all_fueltype_reg_yh)
            result_dict['lf_cy_all_regs_av'] = load_factor_y[fueltype_int]
        else:
            pass

    # Get load factor
    result_dict['lf_cy'] = {}
    result_dict['lf_by'] = {}
    for year, lf_fueltype_regs in reg_load_factor_y.items():
        print("lf_fueltype_regs")
        print(lf_fueltype_regs.shape)
        if year == base_yr:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['lf_by'][reg_geocode] = lf_fueltype_regs[fueltype_int][reg_nr]

        elif year == comparison_year:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['lf_cy'][reg_geocode] = lf_fueltype_regs[fueltype_int][reg_nr]
        else:
            pass

    labels = []

    #Base year
    x_values, y_values = [], []
    x_values_0_quadrant, y_values_0_quadrant = [], []
    x_values_1_quadrant, y_values_1_quadrant = [], []
    x_values_2_quadrant, y_values_2_quadrant = [], []
    x_values_3_quadrant, y_values_3_quadrant = [], []

    for reg_nr, reg_geocode in enumerate(regions):

        # Change in load factor
        lf_change_p = ((100 / result_dict['lf_by'][reg_geocode]) * result_dict['lf_cy'][reg_geocode]) - 100

        # Change in peak h deman
        demand_peak_h_p = ((100 / result_dict['peak_h_demand_by'][reg_geocode]) * result_dict['peak_h_demand_cy'][reg_geocode]) - 100

        # Change in total regional demand
        tot_demand_p = ((100 / result_dict['demand_by'][reg_geocode]) * result_dict['demand_cy'][reg_geocode]) - 100

        x_values.append(lf_change_p)
        #y_values.append(tot_demand_p)
        y_values.append(demand_peak_h_p)

        labels.append(reg_geocode)

        if lf_change_p < 0 and tot_demand_p > 0:
            x_values_0_quadrant.append(lf_change_p)
            y_values_0_quadrant.append(tot_demand_p)
        elif lf_change_p > 0 and tot_demand_p > 0:
            x_values_1_quadrant.append(lf_change_p)
            y_values_1_quadrant.append(tot_demand_p)
        elif lf_change_p > 0 and tot_demand_p < 0:
            x_values_2_quadrant.append(lf_change_p)
            y_values_2_quadrant.append(tot_demand_p)
        else:
            x_values_3_quadrant.append(lf_change_p)
            y_values_3_quadrant.append(tot_demand_p)

    # Add average
    national_tot_cy_p = ((100 / result_dict['lf_by_all_regs_av']) * result_dict['lf_cy_all_regs_av']) - 100
    #national_tot_demand_p = ((100 / result_dict['demand_by_all_regs']) * result_dict['demand_cy_all_regs']) - 100
    national_peak_h_p = ((100 / result_dict['peak_h_demand_by_all_regs']) * result_dict['peak_h_demand_cy_all_regs']) - 100

    x_val_national_lf_demand_cy = [national_tot_cy_p]
    #y_val_national_lf_demand_cy = [national_tot_demand_p]
    y_val_national_lf_demand_cy = [national_peak_h_p]

    # -------------------------------------
    # Plot
    # -------------------------------------
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(9, 8)) #width, height

    ax = fig.add_subplot(1, 1, 1)

    ax.scatter(
        x_values_0_quadrant,
        y_values_0_quadrant,
        alpha=0.6,
        color='rosybrown',
        s=8)
    ax.scatter(
        x_values_1_quadrant,
        y_values_1_quadrant,
        alpha=0.6,
        color='firebrick',
        s=8)
    ax.scatter(
        x_values_2_quadrant,
        y_values_2_quadrant,
        alpha=0.6,
        color='forestgreen',
        s=8)
    ax.scatter(
        x_values_3_quadrant,
        y_values_3_quadrant,
        alpha=0.6,
        color='darkolivegreen',
        s=8)

    # Add average
    ax.scatter(
        x_val_national_lf_demand_cy,
        y_val_national_lf_demand_cy,
        alpha=1.0,
        color='black',
        s=20,
        marker="v",
        linewidth=0.5,
        edgecolor='black',
        label='national')

    # --------
    # Axis
    # --------
    ax.set_xlabel("load factor change (%) {}".format(fueltype_str))
    ax.set_ylabel("change in peak h (%) {}".format(fueltype_str))

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='on',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='on') # labels along the bottom edge are off

    # --------
    # Grd
    # --------
    ax.set_axisbelow(True)
    ax.set_xticks([0], minor=True)
    ax.set_yticks([0], minor=True)
    ax.yaxis.grid(True, which='minor', linewidth=0.7, color='grey', linestyle='--')
    ax.xaxis.grid(True, which='minor', linewidth=0.7, color='grey', linestyle='--')

    # Limit
    #plt.ylim(ymin=0)

    # -----------
    # Labelling
    # -----------
    if label_points:
        for pos, txt in enumerate(labels):
            ax.text(
                x_values[pos],
                y_values[pos],
                txt,
                horizontalalignment="right",
                verticalalignment="top",
                fontsize=6)
    # --------
    # Legend
    # --------
    '''legend(
        loc=3,
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)'''

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()
    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    plt.close()

def plot_cross_graphs_scenarios(
        base_yr,
        comparison_year,
        regions,
        scenario_data,
        fueltype_int,
        fueltype_str,
        fig_name,
        label_points,
        plotshow):

    result_dict = defaultdict(dict)

    # -------------------------------------------
    # Get base year modelled demand of any scenario (becasue base year the same in every scenario)
    # -------------------------------------------$
    all_scenarios = list(scenario_data.keys())
    first_scenario = all_scenarios[0]
    ed_year_fueltype_regs_yh = scenario_data[first_scenario]['ed_fueltype_regs_yh']

    for year, fuels in ed_year_fueltype_regs_yh.items():

        if year == base_yr:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['demand_by'][reg_geocode] = np.sum(fuels[fueltype_int][reg_nr])
                result_dict['peak_h_demand_by'][reg_geocode] = np.max(fuels[fueltype_int][reg_nr])

            # Demand across all regs
            result_dict['demand_by_all_regs'] = np.sum(fuels[fueltype_int])

            # Peak demand
            fuel_all_reg_yh = np.sum(fuels[fueltype_int], axis=0)# Sum across all regions
            result_dict['peak_h_demand_by_all_regs'] = np.max(fuel_all_reg_yh)

            # Calculate national load factor
            fuel_all_fueltype_reg_yh = np.sum(fuels, axis=1)# Sum across all regions per fueltype
            load_factor_y = load_factors.calc_lf_y(fuel_all_fueltype_reg_yh)
            result_dict['lf_by_all_regs_av'] = load_factor_y[fueltype_int]

    for scenario, data in scenario_data.items():
        result_dict['demand_cy'][scenario] = {}
        result_dict['peak_h_demand_cy'][scenario] = {}

        for year, fuels in data['ed_fueltype_regs_yh'].items():
            if year == comparison_year:
                for reg_nr, reg_geocode in enumerate(regions):
                    result_dict['demand_cy'][scenario][reg_geocode] = np.sum(fuels[fueltype_int][reg_nr])
                    result_dict['peak_h_demand_cy'][scenario][reg_geocode] = np.max(fuels[fueltype_int][reg_nr])

                # Demand across all regs
                result_dict['demand_cy_all_regs'][scenario] = np.sum(fuels[fueltype_int])

                # Peak demand
                fuel_all_reg_yh = np.sum(fuels[fueltype_int], axis=0) # Sum across all regions
                result_dict['peak_h_demand_cy_all_regs'][scenario] = np.max(fuel_all_reg_yh)

                # Calculate national load factor
                fuel_all_fueltype_reg_yh = np.sum(fuels, axis=1)# Sum across all regions per fueltype
                load_factor_y = load_factors.calc_lf_y(fuel_all_fueltype_reg_yh)
                result_dict['lf_cy_all_regs_av'][scenario] = load_factor_y[fueltype_int]
            else:
                pass

    # Get load factor of base year
    reg_load_factor_y = scenario_data[first_scenario]['reg_load_factor_y']
    for year, lf_fueltype_regs in reg_load_factor_y.items():
        if year == base_yr:
            for reg_nr, reg_geocode in enumerate(regions):
                result_dict['lf_by'][reg_geocode] = lf_fueltype_regs[fueltype_int][reg_nr]

    for scenario, data in scenario_data.items():
        reg_load_factor_y = scenario_data[scenario]['reg_load_factor_y']
        result_dict['lf_cy'][scenario] = {}

        for year, lf_fueltype_regs in reg_load_factor_y.items():
            if year == comparison_year:
                for reg_nr, reg_geocode in enumerate(regions):
                    result_dict['lf_cy'][scenario][reg_geocode] = lf_fueltype_regs[fueltype_int][reg_nr]
            else:
                pass

    # --------------------------
    # Iterate scenario and plot
    # --------------------------
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(18, 8)) #width, height

    ax = fig.add_subplot(1, 1, 1)

    color_list = plotting_styles.color_list_scenarios()

    marker_list = plotting_styles.marker_list()

    all_x_values = []
    all_y_values = []

    for scenario_nr, scenario in enumerate(all_scenarios):
        labels = []
        x_values, y_values = [], []

        for reg_nr, reg_geocode in enumerate(regions):

            # Change in load factor
            lf_change_p = (
                (100 / result_dict['lf_by'][reg_geocode]) * result_dict['lf_cy'][scenario][reg_geocode]) - 100

            # Change in peak h deman
            demand_peak_h_p = (
                (100 / result_dict['peak_h_demand_by'][reg_geocode]) * result_dict['peak_h_demand_cy'][scenario][reg_geocode]) - 100

            # Change in total regional demand
            tot_demand_p = (
                (100 / result_dict['demand_by'][reg_geocode]) * result_dict['demand_cy'][scenario][reg_geocode]) - 100

            x_values.append(lf_change_p)
            #y_values.append(tot_demand_p)
            y_values.append(demand_peak_h_p)

            #print("ADDING {} {} {}".format(reg_nr, round(lf_change_p, 3), round(demand_peak_h_p, 3)))

            labels.append(reg_geocode)

        # Add average
        national_tot_cy_p = ((100 / result_dict['lf_by_all_regs_av']) * result_dict['lf_cy_all_regs_av'][scenario]) - 100
        #national_tot_demand_p = ((100 / result_dict['demand_by_all_regs']) * result_dict['demand_cy_all_regs'][scenario]) - 100
        national_peak_h_p = ((100 / result_dict['peak_h_demand_by_all_regs']) * result_dict['peak_h_demand_cy_all_regs'][scenario]) - 100

        x_val_national_lf_demand_cy = [national_tot_cy_p]
        #y_val_national_lf_demand_cy = [national_tot_demand_p]
        y_val_national_lf_demand_cy = [national_peak_h_p]

        all_x_values += x_values
        all_y_values += y_values
        # -------------------------------------
        # Plot
        # -------------------------------------
        color = color_list[scenario_nr]
        marker = marker_list[scenario_nr]

        alpha_value = 0.6
        marker_size = 7
        ax.scatter(
            x_values,
            y_values,
            alpha=alpha_value,
            color=color,
            #marker=marker,
            edgecolor=color,
            linewidth=0.5,
            s=marker_size,
            label=scenario)

        # Add average
        ax.scatter(
            x_val_national_lf_demand_cy,
            y_val_national_lf_demand_cy,
            alpha=1.0,
            color=color,
            s=20,
            marker="v",
            linewidth=0.5,
            edgecolor='black',
            label='national')

    # --------
    # Axis
    # --------
    ax.set_xlabel("load factor change (%) {}".format(fueltype_str))
    ax.set_ylabel("change in peak h (%) {}".format(fueltype_str))

    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='on',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='on') # labels along the bottom edge are off

    # --------
    # Grd
    # --------
    ax.set_axisbelow(True)
    #ax.grid(True)
    #ax.set_xticks(minor=False)
    ax.set_xticks([0], minor=True)
    #ax.set_yticks(minor=False)
    ax.set_yticks([0], minor=True)
    #ax.yaxis.grid(True, which='major')
    ax.yaxis.grid(True, which='minor', linewidth=0.7, color='grey', linestyle='--')
    #ax.xaxis.grid(True, which='major')
    ax.xaxis.grid(True, which='minor', linewidth=0.7, color='grey', linestyle='--')

    # Limit
    #plt.ylim(ymin=0)

    # -----------
    # Labelling
    # -----------
    if label_points:
        for pos, txt in enumerate(labels):
            ax.text(
                x_values[pos],
                y_values[pos],
                txt,
                horizontalalignment="right",
                verticalalignment="top",
                fontsize=6)

    # -------
    # Title information
    # -------
    max_lf = round(max(all_x_values), 2)
    min_lf = round(min(all_x_values), 2)
    min_peak_h = round(min(all_y_values), 2)
    max_peak_h = round(max(all_y_values), 2)

    font_additional_info = plotting_styles.font_info(size=4)

    plt.title(
        "max_peak_h: {} min_peak_h: {}, min_lf: {} max_lf: {}".format(
            max_peak_h,
            min_peak_h,
            min_lf,
            max_lf),
        fontdict=font_additional_info)

    # --------
    # Legend
    # --------
    plt.legend(
        loc='best',
        ncol=1,
        prop={
            'family': 'arial',
            'size': 3},
        frameon=False)

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()
    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    plt.close()



"""Testing functions ``basic`` ``date_prop``
"""
from datetime import date
from energy_demand.basic import date_prop

'''def test_get_dates_week_nr():
    """testing function
    """
    import datetime
    from isoweek import Week
    in_year = 2017
    week_nr = 33
    expected = [
        datetime.date(2017, 8, 14),
        datetime.date(2017, 8, 15),
        datetime.date(2017, 8, 16),
        datetime.date(2017, 8, 17),
        datetime.date(2017, 8, 18),
        datetime.date(2017, 8, 19),
        datetime.date(2017, 8, 20)
    ]

    # call function
    out_value = date_prop.get_dates_week_nr(in_year, week_nr)

    assert out_value == expected
'''

def test_get_season():
    """testing
    """
    result = date_prop.get_season(year_to_model=2018)

    assert result['summer'][0] == 151 #2018 regular year, first of june julianday - 1 for python
    assert result['autumn'][0] == 243 #2018 regular year, first of june julianday - 1 for python
    assert result['winter'][0] == 334 #2018 regular year, first of june julianday - 1 for python
    assert result['spring'][0] == 59 #2018 regular year, first of june julianday - 1 for python

def test_get_yeardays_daytype():
    """testing
    """
    model_yeardays_daytype, yeardays_month, yeardays_month_days = date_prop.get_yeardays_daytype(
        year_to_model=2018)

    assert model_yeardays_daytype[0] == 'workday'
    assert model_yeardays_daytype[1] == 'workday'  
    assert model_yeardays_daytype[2] == 'workday'
    assert model_yeardays_daytype[3] == 'workday' 
    assert model_yeardays_daytype[4] == 'workday'
    assert model_yeardays_daytype[5] == 'holiday'
    assert model_yeardays_daytype[6] == 'holiday'

def test_get_month_from_yeraday():
    """testing
    """

    result = date_prop.get_month_from_yeraday(year=2015, yearday=37)
    expected = 1 #Feb
    assert result == expected

def test_date_to_yearday():
    """Testing
    """
    in_year = 2015
    in_month = 6
    in_day = 13
    expected = 164 - 1

    # call function
    out_value = date_prop.date_to_yearday(in_year, in_month, in_day)

    assert out_value == expected

def test_yearday_to_date():
    """Testing
    """
    in_year = 2015
    in_month = 6
    in_day = 13
    in_yearday = 163
    expected = date(2015, in_month, in_day)

    # call function
    out_value = date_prop.yearday_to_date(in_year, in_yearday)

    assert out_value == expected

def test_fullyear_dates():
    """Testing
    """
    start_date = date(2015, 1, 1)
    end_date = date(2015, 1, 4)
    expected = [
        date(2015, 1, 1),
        date(2015, 1, 2),
        date(2015, 1, 3),
        date(2015, 1, 4)]

    # call function
    out_value = date_prop.fullyear_dates(start_date, end_date)

    assert out_value == expected

def test_get_weekday_type():
    """Testing
    """
    date_to_test = date(2015, 4, 3)
    expected = 'holiday'

    # call function
    out_value = date_prop.get_weekday_type(date_to_test)

    assert out_value == expected

    date_to_test = date(2017, 9, 7)
    expected = 'working_day'

    # call function
    out_value = date_prop.get_weekday_type(date_to_test)

    assert out_value == expected

    date_to_test = date(2017, 9, 9)
    expected = 'holiday'

    # call function
    out_value = date_prop.get_weekday_type(date_to_test)

    assert out_value == expected

    # Test year
    date_to_test = date(2015, 4, 3)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2014, 4, 21)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected
    date_to_test = date(2013, 4, 1)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2012, 4, 9)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2011, 1, 2)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected
    
    date_to_test = date(2010, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected
    
    date_to_test = date(2009, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected
    
    date_to_test = date(2008, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected
    
    date_to_test = date(2007, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2006, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2005, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2004, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2003, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected

    date_to_test = date(2002, 12, 25)
    expected = 'holiday'
    out_value = date_prop.get_weekday_type(date_to_test)
    assert out_value == expected



"""Testing
"""

from pkg_resources import Requirement
from pkg_resources import resource_filename
import os

class TestDataPath:
    """Tests that the data_files configuration in setup.cfg places the data
    folder in a location relative to the energy_demand package
    """
    '''def test_get_path(self):
        path_main = resource_filename(Requirement.parse("energy_demand"),
                                      "data")
        # CHECK THAT DATA FOLDER IN NISMOD FOLDER
        expected = os.path.join(path_main)
        assert os.path.exists(expected)

        expected = os.path.join(path_main, 'submodel_residential')
        assert os.path.exists(expected)

        expected = os.path.join(path_main, 'submodel_service')
        assert os.path.exists(expected)'''




"""Testing
"""
import numpy as np
from energy_demand.profiles import load_profile

def test_calc_yh():
    """testing"""
    shape_yd = np.ones((365))
    shape_y_dh = np.zeros((365, 24))

    shape_yd[0] = 2.0 #add plus one
    shape_y_dh[0] = np.ones((24))
    shape_y_dh[0][0] = 2.0 #add plus one in day

    out = load_profile.calc_yh(
        shape_yd=shape_yd,
        shape_y_dh=shape_y_dh,
        model_yeardays=list(range(365)))

    assert out[0][0] == 2 * 2
    assert out[0][1] == 2 * 1

def test_abs_to_rel():
    """Test
    """
    absolute_array = np.array([1, 2, 3])
    absolute_array2 = np.array([0, 0, 0])

    relative_array = load_profile.abs_to_rel(absolute_array)
    relative_array2 = load_profile.abs_to_rel(absolute_array2)

    expected_relative_array = np.array(
        [
            float(absolute_array[0]) / np.sum(absolute_array),
            float(absolute_array[1]) / np.sum(absolute_array),
            float(absolute_array[2]) / np.sum(absolute_array)
            ]
    )

    np.testing.assert_equal(np.round(relative_array,3), np.round(expected_relative_array, 3))
    np.testing.assert_equal(relative_array2, absolute_array2)

def test_calk_peak_h_dh():
    """testing
    """
    fuel_yh = np.zeros((2, 24)) #two fueltypes, 24 hours
    fuel_yh[1] = np.array([1,2,3,4,5,4,3,2,1,1,1,1,1,1,1,1,2,2,1,1,3,4,5,6])

    result = load_profile.calk_peak_h_dh(fuel_yh)

    assert result[0] == np.max(fuel_yh[0])
    assert result[1] == np.max(fuel_yh[1])

def test_LoadProfileStock():
    """testing
    """

    result_obj = load_profile.LoadProfileStock("test_stock")

    expected = "test_stock"

    assert result_obj.name == expected
    model_yeardays = list(range(365))
    # -----
    result_obj.add_lp(
        unique_identifier="A123",
        technologies=['placeholder_tech'],
        enduses=['cooking'],
        shape_y_dh=np.zeros((365)),
        shape_yd=np.zeros((365)),
        shape_yh=np.zeros((365, 24)),
        sectors=[False],
        model_yeardays=model_yeardays)

    result = result_obj.stock_enduses

    result2 = load_profile.get_stock_enduses(result_obj.load_profiles)

    assert result == ['cooking']
    assert result2 == ['cooking']

    # ---
    # -----
    result_obj.add_lp(
        unique_identifier="A123",
        technologies=['placeholder_tech'],
        enduses=['cooking'],
        shape_y_dh=np.zeros((365)),
        shape_yd=np.zeros((365)),
        shape_yh=np.zeros((365, 24)),
        sectors=['sectorA'],
        model_yeardays=model_yeardays)

    result = result_obj.stock_enduses
    assert result == ['cooking']

    # test get_lp()
    np.testing.assert_array_equal(np.zeros((365, 24)), result_obj.get_lp('cooking', 'sectorA', 'placeholder_tech', 'shape_yh'))

def test_generate_key_lu_dict():
    """
    """
    result = load_profile.generate_key_lu_dict(
        {"A": 3},
        "100D",
        ['cooking'],
        ["sectorA", "sectorB"],
        ['techA'])
    
    expected = {"A": 3, ('cooking', "sectorA", 'techA'): "100D", ('cooking', "sectorB", 'techA'): "100D"}

    assert expected == result

def test_LoadProfile():

    result_obj = load_profile.LoadProfile(
        enduses=['heating'],
        unique_identifier="A123",
        shape_yd=np.zeros((365)),
        shape_yh=np.zeros((365, 24)),
        shape_y_dh=np.full((365), 1/365),
        model_yeardays=list(range(365)))

    assert result_obj.enduses == ['heating']

def test_calc_av_lp():

    model_yeardays_daytype = [
        'holiday',
        'workday',
        'workday',
        'workday',

        'holiday',
        'workday',
        'workday',
        'workday',

        'holiday',
        'workday',
        'workday',
        'workday',

        'holiday',
        'workday',
        'workday',
        'workday',
        ] #

    # Example with 16 days
    seasons = {}
    seasons['winter'] = [0, 1, 2, 3]
    seasons['spring'] = [4, 5, 6, 7]
    seasons['summer'] = [8, 9, 10, 11]
    seasons['autumn'] = [12, 13, 14, 15]

    demand_yh = np.zeros((16, 24)) + 10

    for i in range(4):
        demand_yh[i] = 12

    for i in range(4, 8):
        demand_yh[i] = 10

    # weekend
    demand_yh[0] = 100
    demand_yh[4] = 200

    result_av, _ = load_profile.calc_av_lp(demand_yh, seasons, model_yeardays_daytype)
    
    expected = {
        'winter': {'workday': np.full((24), 12), 'holiday': np.full((24), 100)},
        'spring': {'workday': np.full((24), 10), 'holiday': np.full((24), 200)},
        'summer': {'workday': np.full((24), 12), 'holiday': np.full((24), 100)},
        'autmn': {'workday': np.full((24), 10), 'holiday': np.full((24), 10)}
        }

    np.testing.assert_array_equal(expected['winter']['workday'], result_av['winter']['workday'])
    np.testing.assert_array_equal(expected['winter']['holiday'], result_av['winter']['holiday'])
    np.testing.assert_array_equal(expected['spring']['workday'], result_av['spring']['workday'])
    np.testing.assert_array_equal(expected['spring']['holiday'], result_av['spring']['holiday'])



from energy_demand.read_write.narrative_related import crit_dim_var
from collections import OrderedDict

class TestCritDimensionsVar:

    def test_crit_in_list(self):

        crit_in_list = [
            {'sig_midpoint': 0, 'value_by': 5, 'diffusion_choice': 'linear',
            'fueltype_replace': 0, 'regional_specific': False,
            'base_yr': 2015, 'value_ey': 5, 'sig_steepness': 1,
            'end_yr': 2050, 'fueltype_new': 0}]

        actual = crit_dim_var(crit_in_list)
        expected = True
        print("AC ULT " + str(actual))
        assert actual == expected

    def test_nested_dict_in_list(self):

        nested_dict_in_list = [{'key': {'nested_dict': 'a value'}}]

        actual = crit_dim_var(nested_dict_in_list)
        expected = False
        print("AC ULT " + str(actual))
        assert actual == expected

    def test_nested_dict(self):

        nested_dict = {'key': {'nested_dict': 'a value'}}

        actual = crit_dim_var(nested_dict)
        expected = False
        assert actual == expected

        nested_dict = {'another_key': {}, 'key': []}
        actual = crit_dim_var(nested_dict)
        expected = True
        assert actual == expected

    def test_crit_list(self):

        a_list = []
        actual = crit_dim_var(a_list)
        expected = True
        assert actual == expected

        the_bug = {'enduse': [], 'diffusion_type': 'linear', 'default_value': 15.5,
                'name': 'ss_t_base_heating', 'sector': True,
                'regional_specific': False,
                'description': 'Base temperature',
                'scenario_value': 15.5}
        actual = crit_dim_var(the_bug)
        expected = True
        assert actual == expected

    def test_crit_dim_var_ordered(self):

        the_bug = OrderedDict()
        the_bug['a'] = {'diffusion_type': 'linear', 'default_value': 15.5,
                        'name': 'ss_t_base_heating', 'sector': True,
                        'regional_specific': False,
                        'description': 'Base temperature',
                        'scenario_value': 15.5,
                        'enduse': []}
        the_bug['b'] = 'vakye'
        actual = crit_dim_var(the_bug)
        expected = False
        assert actual == expected

    def test_crit_dim_var_buggy(self):

        fixture = [
            {'sig_steepness': 1,
             'sector': 'dummy_sector',
             'diffusion_choice': 'linear',
             'sig_midpoint': 0,
             'regional_specific': True,
             'base_yr': 2015,
             'value_ey': 0.05,
             'value_by': 0.05,
             'regional_vals_ey': {'W06000023': 0.05, 'W06000010': 0.05},
             'regional_vals_by': {'W06000023': 0.05, 'W06000010': 0.05},
             'end_yr': 2030,
             'enduse': [],
             'default_by': 0.05
             },
            {'sig_steepness': 1,
             'sector': 'dummy_sector',
             'diffusion_choice': 'linear',
             'sig_midpoint': 0,
             'regional_specific': True,
             'base_yr': 2030,
             'value_ey': 0.05,
             'value_by': 0.05,
             'regional_vals_ey': {'W06000023': 0.05, 'W06000010': 0.05},
             'regional_vals_by': {'W06000023': 0.05, 'W06000010': 0.05},
             'end_yr': 2050,
             'enduse': [],
             'default_by': 0.05
             }
            ]
        actual = crit_dim_var(fixture)
        expected = True
        assert actual == expected



"""Script containing functions to calculate load factors and
also peak shifting methods which are used to implement
demand management"""
import numpy as np

def peak_shaving_max_min(
        loadfactor_yd_cy_improved,
        average_yd,
        fuel_yh,
        mode_constrained
    ):
    """Shift demand with help of load factor. All demand above
    the maximum load is shifted proportionally to all hours
    having below average demand.

    Arguments
    ----------
    loadfactor_yd_cy_improved : array
        Improved shifted daily load fuel
    average_yd : array
        Average load for every day in year
    fuel_yh : array
        Fuel for every hour
    mode_constrained : bool
        Running mode information

    Returns
    -------
    shifted_fuel_yh : array
        Shifted fuel

    Info
    ----
        Steps:
            - Calculate new maximum demand for every hour and fueltype
            - Calculate difference in demand to mean for every day and fueltype
            - Calculate percentage of demand for every hour with lower demand than average
            - Calculate total demand above average and shift to yh
            - Set all hours with more demand than maximum peak to maximum peak demand
    """
    # ------------------------------------------
    # Calculate new maximum demand for every day
    # and fueltype with help of newly adaped load factor
    # ------------------------------------------
    allowed_demand_max_d = average_yd / loadfactor_yd_cy_improved
    allowed_demand_max_d[np.isnan(allowed_demand_max_d)] = 0

    if mode_constrained:
        average_yd = average_yd[:, np.newaxis]
        allowed_demand_max_d = allowed_demand_max_d[:, np.newaxis]
    else:
        average_yd = average_yd[:, :, np.newaxis]
        allowed_demand_max_d = allowed_demand_max_d[:, :, np.newaxis]

    # ------------------------------------------
    # Calculate difference to daily mean for every hour
    # for every fueltype (hourly value - daily mean)
    # ------------------------------------------
    diff_to_mean = fuel_yh - average_yd

    # ------------------------
    # Calculate areas of lp below average for every day
    # all lp higher than average are set to zero
    # ------------------------
    diff_to_mean[diff_to_mean > 0] = 0
    diff_to_mean = np.abs(diff_to_mean)

    # Sum along all fueltypes the total fuels which are lp below average
    # Calculate percentage of total shiftable from above average to
    # below average for all hours which can take on fuel
    if mode_constrained:
        tot_area_below_mean = np.sum(diff_to_mean, axis=1) #one fueltype
        tot_area_below_mean = tot_area_below_mean[:, np.newaxis]
    else:
        tot_area_below_mean = np.sum(diff_to_mean, axis=2) #multiple fueltypes
        tot_area_below_mean = tot_area_below_mean[:, :, np.newaxis]

    area_below_mean_p = diff_to_mean / tot_area_below_mean
    area_below_mean_p[np.isnan(area_below_mean_p)] = 0

    # Calculate diff to newmax for every hour
    diff_to_max_demand_d = fuel_yh - allowed_demand_max_d
    diff_to_max_demand_d[diff_to_max_demand_d < 0] = 0

    # -----------------------------------------
    # Start with largest deviation to mean
    # and shift to all hours below average
    # -----------------------------------------
    # Calculate total demand which is to be shifted
    if mode_constrained:
        tot_demand_to_shift = np.sum(diff_to_max_demand_d, axis=1) # one fueltype
        tot_demand_to_shift = tot_demand_to_shift[:, np.newaxis]
    else:
        tot_demand_to_shift = np.sum(diff_to_max_demand_d, axis=2) # multiple fueltypes
        tot_demand_to_shift = tot_demand_to_shift[:, :, np.newaxis]

    # Add fuel below average:
    # Distribute shiftable demand to all hours which are below average
    # according to percentage contributing to lf which is below average
    shifted_fuel_yh = fuel_yh + (area_below_mean_p * tot_demand_to_shift)

    # Set all fuel hours whih are above max to max (substract diff)
    shifted_fuel_yh = shifted_fuel_yh - diff_to_max_demand_d

    #assert np.sum(shifted_fuel_yh) == np.sum(fuel_yh)
    return shifted_fuel_yh

def calc_lf_y(fuel_yh):
    """Calculate the yearly load factor for every fueltype
    by dividing the yearly average load by the peak hourly
    load in a year.

    Arguments
    ---------
    fuel_yh : array (fueltypes, 365, 24) or (fueltypes, 8760)
        Yh fuel

    Returns
    -------
    load_factor_y : array
        Yearly load factors as percentage (100% = 1)

    Note
    -----
    Load factor = average load / peak load
    https://en.wikipedia.org/wiki/Load_factor_(electrical)
    https://circuitglobe.com/load-factor.html
    """
    if fuel_yh.shape[1] == 365:
        fuel_yh_8760 = fuel_yh.reshape(fuel_yh.shape[0], 8760)
    else:
        fuel_yh_8760 = fuel_yh

    # Get total sum per fueltype
    tot_load_y = np.sum(fuel_yh_8760, axis=1)

    # Calculate maximum hour in every day of a year
    max_load_h = np.max(fuel_yh_8760, axis=1)

    # Caclualte yearly load factor for every fueltype
    with np.errstate(divide='ignore', invalid='ignore'):
        load_factor_y = (tot_load_y / 8760) / max_load_h

    load_factor_y[np.isnan(load_factor_y)] = 0

    return load_factor_y

def calc_lf_y_8760(fuel_yh_8760):
    """Calculate the yearly load factor for every fueltype
    by dividing the yearly average load by the peak hourly
    load in a year.

    Arguments
    ---------
    fuel_yh_8760 : array
        Fueltypes, regions, 8760hours

    Returns
    -------
    load_factor_y : array
        Yearly load factors as percentage (100% = 1)

    Note
    -----
    Load factor = average load / peak load
    https://en.wikipedia.org/wiki/Load_factor_(electrical)
    https://circuitglobe.com/load-factor.html
    """
    # Get total sum per fueltype
    tot_load_y = np.sum(fuel_yh_8760)

    # Calculate maximum hour in every day of a year
    max_load_h = np.max(fuel_yh_8760)

    # Caclualte yearly load factor for every fueltype
    with np.errstate(divide='ignore', invalid='ignore'):
        load_factor_y = (tot_load_y / 8760) / max_load_h

    if np.isnan(load_factor_y):
        load_factor_y = 0

    return load_factor_y

def calc_lf_season(seasons, fuel_region_yh, average_fuel_yd):
    """Calculate load factors per fueltype per region.
    The load factor is calculated based on average
    yearly load and maximum saisonal peak factor.

    Arguments
    ---------
    seasons : dict
        Seasons containing yeardays for four seasons
    fuel_region_yh : array
        Fuels
    average_fuel_yd : array
        Average fuels

    Returns
    -------
    seasons_lfs : dict
        Load factors per fueltype and season

    Note
    ----
    If not the yearly average is used for calculation,
    only the load factors within the regions are calculated.
    """
    seasons_lfs = {}
    for season, yeardays_modelled in seasons.items():

        average_fuel_yd_full_year = np.average(
            average_fuel_yd[:, ],
            axis=1)

        # Calculate maximum hour in year
        max_load_h_days_season = np.max(
            fuel_region_yh[:, yeardays_modelled],
            axis=2)

        max_load_h_season = np.max(max_load_h_days_season, axis=1)

        # Unable local RuntimeWarning: divide by zero encountered
        with np.errstate(divide='ignore', invalid='ignore'):

            #convert to decimal
            season_lf = (average_fuel_yd_full_year / max_load_h_season)

        # Replace
        season_lf[np.isinf(season_lf)] = 0
        season_lf[np.isnan(season_lf)] = 0

        seasons_lfs[season] = season_lf

    return seasons_lfs

def calc_lf_d(fuel_yh, average_fuel_yd, mode_constrained):
    """Calculate the daily load factor for every day in a year
    by dividing for each day the daily average by the daily peak
    hour load. The load factor is given in %

    Arguments
    ---------
    fuel_yh : array
        Fuel for every hour in a year
    average_fuel_yd : array
        Average load per day
    mode_constrained : bool
        Mode information

    Returns
    -------
    daily_lf : array
        Laod factor calculated for every modelled day [in %]
    average_fuel_yd : array
        Average fuel for every day
    """
    # Get maximum hours in every day
    if mode_constrained:
        max_load_yd = np.max(fuel_yh, axis=1) #single fueltype
    else:
        max_load_yd = np.max(fuel_yh, axis=2) #multiple fueltypes

    # Unable local RuntimeWarning: divide by zero encountered
    with np.errstate(divide='ignore', invalid='ignore'):

        #convert to decimal
        daily_lf = (average_fuel_yd / max_load_yd)

    # Replace by zero
    daily_lf[np.isinf(daily_lf)] = 0
    daily_lf[np.isnan(daily_lf)] = 0

    return daily_lf

def calc_lf_d_8760(fuel_yh):
    """Calculate the daily load factor for every day in a year
    by dividing for each day the daily average by the daily peak
    hour load. The load factor is given in %

    Arguments
    ---------
    fuel_yh : array
        Fuel for every hour in a year
    average_fuel_yd : array
        Average load per day
    mode_constrained : bool
        Mode information

    Returns
    -------
    daily_lf : array
        Laod factor calculated for every modelled day [in %]
    average_fuel_yd : array
        Average fuel for every day
    """
    # Get maximum hours in every day
    fuel_yh = fuel_yh.reshape(365, 24)

    max_load_yd = np.max(fuel_yh, axis=1)
    average_fuel_yd = np.average(fuel_yh, axis=1)

    # Unable local RuntimeWarning: divide by zero encountered
    with np.errstate(divide='ignore', invalid='ignore'):

        #convert to decimal
        daily_lf = (average_fuel_yd / max_load_yd)

    # Replace by zero
    daily_lf[np.isinf(daily_lf)] = 0
    daily_lf[np.isnan(daily_lf)] = 0

    return daily_lf

def calc_lf_season_8760(seasons, fuel_region_yh):
    """Calculate load factors per fueltype per region.
    The load factor is calculated based on average
    yearly load and maximum saisonal peak factor.

    Arguments
    ---------
    seasons : dict
        Seasons containing yeardays for four seasons
    fuel_region_yh : array
        Fuels

    Returns
    -------
    seasons_lfs : dict
        Load factors per fueltype and season

    Note
    ----
    If not the yearly average is used for calculation,
    only the load factors within the regions are calculated.
    """
    fuel_region_yh_365_24 = fuel_region_yh.reshape(365, 24)
    seasons_lfs = {}

    for season, yeardays_modelled in seasons.items():

        average_fuel_yd = np.average(fuel_region_yh_365_24, axis=1)

        # Calculate maximum hour in year
        max_load_h_days_season = np.max(
            fuel_region_yh_365_24[yeardays_modelled], axis=1)

        max_load_h_season = np.max(max_load_h_days_season)

        # Unable local RuntimeWarning: divide by zero encountered
        with np.errstate(divide='ignore', invalid='ignore'):
            season_lf = (average_fuel_yd / max_load_h_season)

        # Replace
        season_lf[np.isinf(season_lf)] = 0
        season_lf[np.isnan(season_lf)] = 0

        seasons_lfs[season] = season_lf

    return seasons_lfs



"""Compare gas/elec demand on Local Authority Districts with modelled demand
"""
import os
import operator
import logging
import copy
import numpy as np
import matplotlib.pyplot as plt
from scipy import stats

from energy_demand.basic import lookup_tables
from energy_demand import enduse_func
from energy_demand.profiles import load_profile
from energy_demand.validation import elec_national_data
from energy_demand.read_write import data_loader
from energy_demand.basic import date_prop
from energy_demand.plotting import fig_load_profile_dh_multiple
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import plotting_styles

def map_LAD_2011_2015(lad_data):
    """Map LAD 2015 values to LAD 2011.

    Arguments
    -----------
    lad_data : dict
        LAD 2015 data

    Returns
    --------
    mapped_lads : dict
        LAD 2011 census data lads
    """
    mapped_lads = copy.deepcopy(lad_data)

    mapped_lads.keys()

    try:
        # E41000324 (City of London, Westminster) splits
        # to E09000001 (City of London) and E09000033 (Westminster)
        mapped_lads['E41000324'] = lad_data['E09000001'] + lad_data['E09000033']
        del mapped_lads['E09000001']
        del mapped_lads['E09000033']
    except:
        pass

    try:
        # E41000052 (Cornwall, Isles of Scilly) splits
        # to E06000052 (Cornwall) and E06000053 (Isles of Scilly) (edited)
        mapped_lads['E41000052'] = lad_data['E06000052'] + lad_data['E06000053']
        del mapped_lads['E06000052']
        del mapped_lads['E06000053']
    except:
        pass

    try:
        # missing S12000013 (Na h-Eileanan Siar)
        # and S12000027 (Shetland Islands)
        del mapped_lads['S12000013']
        del mapped_lads['S12000027']
    except:
        pass

    return mapped_lads

def temporal_validation(
        result_paths,
        ed_fueltype_yh,
        elec_factored_yh,
        plot_criteria
    ):
    """National hourly electricity data is validated with
    the summed modelled hourly demand for all regions.
    Because the total annual modelled and real demands
    do not match (because of different data sources
    and because Northern Ireland is not included in the
    validation data) a correction factor is used.

    Arguments
    ---------
    result_paths : dict
        Paths
    ed_fueltype_yh : array
        Fuel type specific yh energy demand
    plot_criteria : bool
        Criteria to show plots or not
    """
    # ----------------
    # Plot a full year
    # ----------------
    days_to_plot = list(range(0, 365))
    elec_national_data.compare_results(
        'validation_temporal_electricity_8760h.pdf',
        result_paths['data_results_validation'],
        elec_factored_yh,
        ed_fueltype_yh,
        'all_submodels',
        days_to_plot,
        plot_crit=plot_criteria)

    # Plot four weeks (one of each season)
    winter_week = list(range(
        date_prop.date_to_yearday(2015, 1, 12), date_prop.date_to_yearday(2015, 1, 19))) #Jan
    spring_week = list(range(
        date_prop.date_to_yearday(2015, 5, 11), date_prop.date_to_yearday(2015, 5, 18))) #May
    summer_week = list(range(
        date_prop.date_to_yearday(2015, 7, 13), date_prop.date_to_yearday(2015, 7, 20))) #Jul
    autumn_week = list(range(
        date_prop.date_to_yearday(2015, 10, 12), date_prop.date_to_yearday(2015, 10, 19))) #Oct
    days_to_plot = winter_week + spring_week + summer_week + autumn_week

    elec_national_data.compare_results(
        'validation_temporal_electricity_weeks_selection.pdf',
        result_paths['data_results_validation'],
        elec_factored_yh,
        ed_fueltype_yh,
        'all_submodels',
        days_to_plot,
        plot_crit=plot_criteria)

    return

def spatial_validation_lad_level(
        disaggregated_fuel,
        data_results_validation,
        paths,
        regions,
        reg_coord,
        plot_crit
    ):
    """Spatial validation
    """
    fuel_elec_regs_yh = {}
    fuel_gas_regs_yh = {}
    fuel_gas_residential_regs_yh = {}
    fuel_gas_non_residential_regs_yh = {}
    fuel_elec_residential_regs_yh = {}
    fuel_elec_non_residential_regs_yh = {}

    lookups =  lookup_tables.basic_lookups()
    # -------------------------------------------
    # Spatial validation
    # -------------------------------------------
    subnational_elec = data_loader.read_lad_demands(paths['val_subnational_elec'])
    subnational_elec_residential = data_loader.read_lad_demands(paths['val_subnational_elec_residential'])
    subnational_elec_non_residential = data_loader.read_lad_demands(paths['val_subnational_elec_non_residential'])
    subnational_gas = data_loader.read_lad_demands(paths['val_subnational_gas'])
    subnational_gas_residential = data_loader.read_lad_demands(paths['val_subnational_gas_residential'])
    subnational_gas_non_residential = data_loader.read_lad_demands(paths['val_subnational_gas_non_residential'])
    logging.info("compare total II {}  {}".format(sum(subnational_gas.values()), sum(subnational_gas_residential.values())))

    # Create fueltype secific dict
    for region in regions:
        fuel_elec_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs'][region][lookups['fueltypes']['electricity']]
        fuel_elec_residential_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs_residenital'][region][lookups['fueltypes']['electricity']]
        fuel_elec_non_residential_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs_non_residential'][region][lookups['fueltypes']['electricity']]    
        fuel_gas_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs'][region][lookups['fueltypes']['gas']]
        fuel_gas_residential_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs_residenital'][region][lookups['fueltypes']['gas']]
        fuel_gas_non_residential_regs_yh[region] = disaggregated_fuel['tot_disaggregated_regs_non_residential'][region][lookups['fueltypes']['gas']]

    # ----------------------------------------
    # Remap demands between 2011 and 2015 LADs
    # ----------------------------------------
    subnational_elec = map_LAD_2011_2015(subnational_elec)
    subnational_elec_residential = map_LAD_2011_2015(subnational_elec_residential)
    subnational_elec_non_residential = map_LAD_2011_2015(subnational_elec_non_residential)

    subnational_gas = map_LAD_2011_2015(subnational_gas)
    subnational_gas_residential = map_LAD_2011_2015(subnational_gas_residential)
    subnational_gas_non_residential = map_LAD_2011_2015(subnational_gas_non_residential)

    fuel_elec_regs_yh = map_LAD_2011_2015(fuel_elec_regs_yh)
    fuel_elec_residential_regs_yh = map_LAD_2011_2015(fuel_elec_residential_regs_yh)
    fuel_elec_non_residential_regs_yh = map_LAD_2011_2015(fuel_elec_non_residential_regs_yh)

    fuel_gas_regs_yh = map_LAD_2011_2015(fuel_gas_regs_yh)
    fuel_gas_residential_regs_yh = map_LAD_2011_2015(fuel_gas_residential_regs_yh)
    fuel_gas_non_residential_regs_yh = map_LAD_2011_2015(fuel_gas_non_residential_regs_yh)

    logging.info("compare total {}  {}".format(
        sum(fuel_gas_residential_regs_yh.values()), sum(fuel_gas_regs_yh.values())))

    # --------------------------------------------
    # Correct REAL Values that sum is the same
    # ----------------------------------------------
    data_inputlist = [
        (fuel_elec_residential_regs_yh, subnational_elec_residential),          # domestic
        (fuel_elec_non_residential_regs_yh, subnational_elec_non_residential)]  # nondomestics

    spatial_validation_multiple(
        reg_coord=reg_coord,
        input_data=data_inputlist,
        regions=regions,
        fueltype_str='elec',
        fig_name=os.path.join(data_results_validation, 'validation_multiple_elec.pdf'),
        label_points=False,
        plotshow=plot_crit)

    data_inputlist = [
        (fuel_gas_residential_regs_yh, subnational_gas_residential),          # domestic
        (fuel_gas_non_residential_regs_yh, subnational_gas_non_residential)]  # nondomestics

    spatial_validation_multiple(
        reg_coord=reg_coord,
        input_data=data_inputlist,
        regions=regions,
        fueltype_str='gas',
        fig_name=os.path.join(data_results_validation, 'validation_multiple_gas.pdf'),
        label_points=False,
        plotshow=plot_crit)

    logging.info("... Validation of electricity")
    spatial_validation(
        fuel_elec_regs_yh,
        subnational_elec,
        regions,
        'elec',
        os.path.join(data_results_validation, 'validation_spatial_elec.pdf'),
        label_points=True,
        plotshow=plot_crit)

    logging.info("... Validation of residential electricity")
    spatial_validation(
        fuel_elec_residential_regs_yh,
        subnational_elec_residential,
        regions,
        'elec',
        os.path.join(data_results_validation, 'validation_spatial_residential_elec.pdf'),
        label_points=True,
        plotshow=plot_crit)

    logging.info("... Validation of non-residential electricity")
    spatial_validation(
        fuel_elec_non_residential_regs_yh,
        subnational_elec_non_residential,
        regions,
        'elec',
        os.path.join(data_results_validation, 'validation_spatial_non_residential_elec.pdf'),
        label_points=True,
        plotshow=plot_crit)

    logging.info("... Validation of gas")
    spatial_validation(
        fuel_gas_regs_yh,
        subnational_gas,
        regions,
        'gas',
        os.path.join(data_results_validation, 'validation_spatial_gas.pdf'),
        label_points=True,
        plotshow=plot_crit)

    logging.info("... Validation of residential gas")
    spatial_validation(
        fuel_gas_residential_regs_yh,
        subnational_gas_residential,
        regions,
        'gas',
        os.path.join(data_results_validation, 'validation_spatial_residential_gas.pdf'),
        label_points=True,
        plotshow=plot_crit)

    logging.info("... Validation of non residential gas")
    spatial_validation(
        fuel_gas_non_residential_regs_yh,
        subnational_gas_non_residential,
        regions,
        'gas',
        os.path.join(data_results_validation, 'validation_spatial_non_residential_gas.pdf'),
        label_points=True,
        plotshow=plot_crit)

    return

def temporal_validation_msoa_lad(
        ed_fueltype_national_yh,
        ed_fueltype_regs_yh,
        fueltypes,
        result_paths,
        paths,
        regions,
        reg_coord,
        seasons,
        model_yeardays_daytype,
        plot_crit
    ):
    """Validate national hourly demand for yearls fuel
    for all LADs. Test how the national disaggregation
    works.

    Info
    -----
    Because the floor area is only availabe for LADs from 2001,
    the LADs are converted to 2015 LADs.
    """
    logging.info("... temporal validation")

    # -------------------------------------------
    # Electrictiy demands
    # -------------------------------------------

    # LAD level
    subnational_elec_lad = data_loader.read_lad_demands(
        paths['val_subnational_elec'])

    # MSOA level
    subnational_elec_msoa = data_loader.read_elec_data_msoa(
        paths['val_subnational_msoa_elec'])

    # Create fueltype secific dict
    fuel_elec_regs_yh = {}
    for region_array_nr, region in enumerate(regions):
        gwh_modelled = np.sum(ed_fueltype_regs_yh[fueltypes['electricity']][region_array_nr])
        fuel_elec_regs_yh[region] = gwh_modelled

    # Create fueltype secific dict
    fuel_gas_regs_yh = {}
    for region_array_nr, region in enumerate(regions):
        gwh_modelled = np.sum(ed_fueltype_regs_yh[fueltypes['gas']][region_array_nr])
        fuel_gas_regs_yh[region] = gwh_modelled

    # ----------------------------------------
    # Remap demands between 2011 and 2015 LADs
    # ----------------------------------------
    subnational_elec = map_LAD_2011_2015(subnational_elec_lad)
    fuel_elec_regs_yh = map_LAD_2011_2015(fuel_elec_regs_yh)

    spatial_validation(
        fuel_elec_regs_yh,
        subnational_elec,
        regions,
        'elec',
        os.path.join(result_paths['data_results_validation'], 'validation_spatial_elec_msoa_lad.pdf'),
        label_points=False,
        plotshow=plot_crit)

    return

def spatio_temporal_val(
        ed_fueltype_national_yh,
        ed_fueltype_regs_yh,
        result_paths,
        paths,
        regions,
        seasons,
        model_yeardays_daytype,
        plot_crit
    ):
    """Validate spatial and temporal energy demands

    Info
    -----
    Because the floor area is only availabe for LADs from 2001,
    the LADs are converted to 2015 LADs.
    """
    logging.info("... temporal validation")

    fueltypes =  lookup_tables.basic_lookups()['fueltypes']
    # -------------------------------------------
    # Spatial validation after calculations
    # -------------------------------------------
    subnational_elec = data_loader.read_lad_demands(
        paths['val_subnational_elec'])
    subnational_gas = data_loader.read_lad_demands(
        paths['val_subnational_gas'])

    # Create fueltype secific dict
    fuel_elec_regs_yh = {}
    for region_array_nr, region in enumerate(regions):
        fuel_elec_regs_yh[region] = np.sum(ed_fueltype_regs_yh[fueltypes['electricity']][region_array_nr])

    # Create fueltype secific dict
    fuel_gas_regs_yh = {}
    for region_array_nr, region in enumerate(regions):
        fuel_gas_regs_yh[region] = np.sum(ed_fueltype_regs_yh[fueltypes['gas']][region_array_nr])

    # ----------------------------------------
    # Remap demands between 2011 and 2015 LADs
    # ----------------------------------------
    subnational_elec = map_LAD_2011_2015(subnational_elec)
    subnational_gas = map_LAD_2011_2015(subnational_gas)
    fuel_elec_regs_yh = map_LAD_2011_2015(fuel_elec_regs_yh)
    fuel_gas_regs_yh = map_LAD_2011_2015(fuel_gas_regs_yh)

    spatial_validation(
        fuel_elec_regs_yh,
        subnational_elec,
        regions,
        'elec',
        fig_name=os.path.join(result_paths['data_results_validation'], 'validation_spatial_elec_post_calcualtion.pdf'),
        label_points=False,
        plotshow=plot_crit)

    spatial_validation(
        fuel_gas_regs_yh,
        subnational_gas,
        regions,
        'gas',
        fig_name=os.path.join(result_paths['data_results_validation'], 'validation_spatial_gas_post_calcualtion.pdf'),
        label_points=False,
        plotshow=plot_crit)

    # -------------------------------------------
    # Temporal validation (hourly for national)
    # -------------------------------------------
    # Read validation data
    elec_2015_indo, elec_2015_itsdo = elec_national_data.read_raw_elec_2015(
        paths['val_nat_elec_data'])

    f_diff_elec = np.sum(ed_fueltype_national_yh[fueltypes['electricity']]) / np.sum(elec_2015_indo)
    logging.info("... ed diff modellend and real [p] %s: ", (1 - f_diff_elec) * 100)

    elec_factored_yh = f_diff_elec * elec_2015_indo

    temporal_validation(
        result_paths,
        ed_fueltype_national_yh[fueltypes['electricity']],
        elec_factored_yh,
        plot_crit)

    # ---------------------------------------------------
    # Calculate average season and daytypes and plot
    # ---------------------------------------------------
    logging.info("...calculate average data and plot per season and fueltype")

    calc_av_lp_modelled, calc_lp_modelled = load_profile.calc_av_lp(
        ed_fueltype_national_yh[fueltypes['electricity']],
        seasons,
        model_yeardays_daytype)

    calc_av_lp_real, calc_lp_real = load_profile.calc_av_lp(
        elec_factored_yh,
        seasons,
        model_yeardays_daytype)

    # Plot average daily loads
    fig_load_profile_dh_multiple.run(
        path_fig_folder=result_paths['data_results_validation'],
        path_plot_fig=os.path.join(
            result_paths['data_results_validation'],
            'validation_all_season_daytypes.pdf'),
        calc_av_lp_modelled=calc_av_lp_modelled,
        calc_av_lp_real=calc_av_lp_real,
        calc_lp_modelled=calc_lp_modelled,
        calc_lp_real=calc_lp_real,
        plot_peak=True,
        plot_radar=False,
        plot_all_entries=False,
        plot_max_min_polygon=True,
        plotshow=False,
        max_y_to_plot=60,
        fueltype_str=False,
        year=False)

    # ---------------------------------------------------
    # Validation of national electrictiy demand for peak
    # ---------------------------------------------------
    logging.debug(
        "...validation of peak data: compare peak with data")

    # Because the coldest day is not the same for every region,
    # the coldest day needs to be defined manually or defined
    # by getting the hours with maximum electricity demand

    # Peak across all fueltypes WARNING: Fueltype specific
    peak_day_all_fueltypes = enduse_func.get_peak_day_all_fueltypes(ed_fueltype_national_yh)
    logging.info("Peak day 'peak_day_all_fueltypes': " + str(peak_day_all_fueltypes))

    fueltype = fueltypes['electricity']
    peak_day_electricity, _ = enduse_func.get_peak_day_single_fueltype(ed_fueltype_national_yh[fueltype])
    logging.info("Peak day 'peak_day_electricity': " + str(peak_day_electricity))

    elec_national_data.compare_peak(
        "validation_peak_elec_day_all_fueltypes.pdf",
        result_paths['data_results_validation'],
        elec_2015_indo[peak_day_all_fueltypes],
        ed_fueltype_national_yh[fueltypes['electricity']][peak_day_all_fueltypes],
        peak_day_all_fueltypes)

    elec_national_data.compare_peak(
        "validation_peak_elec_day_only_electricity.pdf",
        result_paths['data_results_validation'],
        elec_2015_indo[peak_day_electricity],
        ed_fueltype_national_yh[fueltypes['electricity']][peak_day_electricity],
        peak_day_electricity)

    # Manual peak day
    peak_day = 19
    elec_national_data.compare_peak(
        "validation_elec_peak_day_{}.pdf".format(peak_day),
        result_paths['data_results_validation'],
        elec_factored_yh[peak_day],
        ed_fueltype_national_yh[fueltypes['electricity']][peak_day],
        peak_day)

    peak_day = 33
    elec_national_data.compare_peak(
        "validation_elec_peak_day_{}.pdf".format(peak_day),
        result_paths['data_results_validation'],
        elec_factored_yh[peak_day],
        ed_fueltype_national_yh[fueltypes['electricity']][peak_day],
        peak_day)

    peak_day_real_electricity, _ = enduse_func.get_peak_day_single_fueltype(elec_2015_indo)
    logging.info("Peak day 'peak_day_electricity': " + str(peak_day_real_electricity))

    #raise Exception
    elec_national_data.compare_peak(
        "validation_elec_peak_day_{}.pdf".format(peak_day_real_electricity),
        result_paths['data_results_validation'],
        elec_factored_yh[peak_day],
        ed_fueltype_national_yh[fueltypes['electricity']][peak_day_real_electricity],
        peak_day_real_electricity)

    # ---------------------------------------------------
    # Validate boxplots for every hour (temporal validation)
    # ---------------------------------------------------
    elec_national_data.compare_results_hour_boxplots(
        "validation_hourly_boxplots_electricity_01.pdf",
        result_paths['data_results_validation'],
        elec_2015_indo,
        ed_fueltype_national_yh[fueltypes['electricity']])

    return

def spatial_validation(
        subnational_modelled,
        subnational_real,
        regions,
        fueltype_str,
        fig_name,
        label_points=False,
        plotshow=False
    ):
    """Compare gas/elec demand for LADs

    Arguments
    ----------
    lad_infos_shapefile : dict
        Infos of shapefile (dbf / csv)
    ed_fueltype_regs_yh : object
        Regional fuel Given as GWh (?)
    subnational_real : dict
        for electricity: Sub-national electrcity demand given as GWh

    Note
    -----
    SOURCE OF LADS:
        - Data for northern ireland is not included in that, however in BEIS dataset!
    """
    logging.debug("... Validation of spatial disaggregation")
    result_dict = {}
    result_dict['real_demand'] = {}
    result_dict['modelled_demand'] = {}
    diff_real_modelled_p = []
    diff_real_modelled_abs = []

    # -------------------------------------------
    # Match ECUK sub-regional demand with geocode and calculate statistics
    # -------------------------------------------
    for region in regions:
        try:
            if subnational_real[region] == 0:
                pass
            else:
                try:
                    real = subnational_real[region]
                    modelled = subnational_modelled[region]

                    # --Sub Regional Electricity demand (as GWh)
                    result_dict['real_demand'][region] = real
                    result_dict['modelled_demand'][region] = modelled

                    diff_real_modelled_p.append(abs(100 - ((100 / real) * modelled)))
                    diff_real_modelled_abs.append(real - modelled)
                except KeyError:
                    pass #not both data for reald and modelled

        except KeyError:
            logging.debug(
                "Sub-national spatial validation: No fuel for region %s", region)

    # Calculate the average deviation between reald and modelled
    av_deviation_real_modelled = np.average(diff_real_modelled_p)
    median_absolute_deviation = np.median(diff_real_modelled_p)    # median deviation

    # Calculate standard deviation
    std_dev_p = np.std(diff_real_modelled_p)        # Given as percent
    std_dev_abs = np.std(diff_real_modelled_abs)    # Given as energy unit

    # -----------------
    # Sort results according to size
    # -----------------
    sorted_dict_real = sorted(
        result_dict['real_demand'].items(),
        key=operator.itemgetter(1))

    # -------------------------------------
    # Plot
    # -------------------------------------
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(9, 8))

    ax = fig.add_subplot(1, 1, 1)

    x_values = np.arange(0, len(sorted_dict_real), 1)

    y_real_demand = []
    y_modelled_demand = []

    labels = []
    for sorted_region in sorted_dict_real:
        geocode_lad = sorted_region[0]

        y_real_demand.append(result_dict['real_demand'][geocode_lad])
        y_modelled_demand.append(result_dict['modelled_demand'][geocode_lad])

        logging.debug(
            "validation %s LAD %s: real: %s modelled: %s  modelled percentage: %s (%sp diff)",
            fueltype_str,
            geocode_lad,
            round(result_dict['real_demand'][geocode_lad], 4),
            round(result_dict['modelled_demand'][geocode_lad], 4),
            round(100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad], 4),
            round(100 - (100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad]), 4))

        labels.append(geocode_lad)

    # Calculate r_squared
    _slope, _intercept, r_value, _p_value, _std_err = stats.linregress(
        y_real_demand,
        y_modelled_demand)

    # --------
    # Axis
    # --------
    plt.tick_params(
        axis='x',          # changes apply to the x-axis
        which='both',      # both major and minor ticks are affected
        bottom='off',      # ticks along the bottom edge are off
        top='off',         # ticks along the top edge are off
        labelbottom='off') # labels along the bottom edge are off

    # ----------------------------------------------
    # Plot
    # ----------------------------------------------
    plt.plot(
        x_values,
        y_real_demand,
        linestyle='None',
        marker='o',
        alpha=0.6,
        markersize=1.6,
        fillstyle='full',
        markerfacecolor='grey',
        markeredgewidth=0.2,
        color='black',
        label='actual')

    plt.plot(
        x_values,
        y_modelled_demand,
        marker='o',
        linestyle='None',
        markersize=1.6,
        alpha=0.6,
        markerfacecolor='white',
        fillstyle='none',
        markeredgewidth=0.5,
        markeredgecolor='blue',
        color='black',
        label='model')

    plt.ylim(ymin=0)

    # -----------
    # Labelling
    # -----------
    if label_points:
        for pos, txt in enumerate(labels):
            ax.text(
                x_values[pos],
                y_modelled_demand[pos],
                txt,
                horizontalalignment="right",
                verticalalignment="top",
                fontsize=1)

    font_additional_info = plotting_styles.font_info(size=4)

    title_info = ('R_2: {}, std_%: {} (GWh {}), av_diff_%: {} median_abs_dev: {}'.format(
        round(r_value, 2),
        round(std_dev_p, 2),
        round(std_dev_abs, 2),
        round(av_deviation_real_modelled, 2),
        round(median_absolute_deviation, 2)))

    plt.title(
        title_info,
        loc='left',
        fontdict=font_additional_info)

    plt.xlabel("UK regions (excluding northern ireland)")
    plt.ylabel("{} [GWh]".format(fueltype_str))

    # --------
    # Legend
    # --------
    plt.legend(
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()
    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    else:
        plt.close()

def spatial_validation_multiple(
        reg_coord,
        input_data,
        regions,
        fueltype_str,
        fig_name,
        label_points=False,
        plotshow=False
    ):
    """Compare gas/elec demand for LADs

    Arguments
    ----------
    lad_infos_shapefile : dict
        Infos of shapefile (dbf / csv)
    ed_fueltype_regs_yh : object
        Regional fuel Given as GWh (?)
    subnational_real : dict
        for electricity: Sub-national electrcity demand given as GWh

    Note
    -----
    SOURCE OF LADS:
        - Data for northern ireland is not included in that, however in BEIS dataset!
    """
    logging.debug("... Validation of spatial disaggregation")

    color_list = ['firebrick', 'darkseagreen']
    label_list = ['domestic', 'non_domestic']

    # -------------------------------------
    # Plot
    # -------------------------------------
    fig = plt.figure(
        figsize=basic_plot_functions.cm2inch(9, 8)) #width, height

    ax = fig.add_subplot(1, 1, 1)

    cnt_color = 0
    for i in input_data:
        subnational_modelled = i[0]
        subnational_real = i[1]

        result_dict = {}
        result_dict['real_demand'] = {}
        result_dict['modelled_demand'] = {}

        # -------------------------------------------
        # Match ECUK sub-regional demand with geocode
        # -------------------------------------------
        for region in regions:
            for reg_geocode in reg_coord:
                if reg_geocode == region:
                    try:
                        # Test wheter data is provided for LAD or owtherwise ignore
                        if subnational_real[reg_geocode] == 0:
                            pass
                        else:
                            # --Sub Regional Electricity demand (as GWh)
                            result_dict['real_demand'][reg_geocode] = subnational_real[reg_geocode]
                            result_dict['modelled_demand'][reg_geocode] = subnational_modelled[reg_geocode]
                    except KeyError:
                        logging.debug(
                            "Sub-national spatial validation: No fuel for region %s", reg_geocode)

        # --------------------
        # Calculate statistics
        # --------------------
        diff_real_modelled_p = []
        diff_real_modelled_abs = []

        y_real_demand = []
        y_modelled_demand = []

        # -----------------
        # Sort results according to size
        # -----------------
        sorted_dict_real = sorted(
            result_dict['real_demand'].items(),
            key=operator.itemgetter(1))
    
        #for reg_geocode in regions:
        for reg_geocode, _ in sorted_dict_real:
            # Test if real and modelled data are both available
            try:
                real = result_dict['real_demand'][reg_geocode]
                modelled = result_dict['modelled_demand'][reg_geocode]

                diff_real_modelled_p.append(abs(100 - ((100 / real) * modelled))) # Average abs deviation
                diff_real_modelled_abs.append(real - modelled)

                y_real_demand.append(real)
                y_modelled_demand.append(modelled)
            except KeyError:
                pass

        # Calculate the average deviation between reald and modelled
        av_deviation_real_modelled = np.average(diff_real_modelled_p)   # average deviation
        median_absolute_deviation = np.median(diff_real_modelled_p)     # median deviation

        # Calculate standard deviation
        std_dev_p = np.std(diff_real_modelled_p)        # Given as percent
        std_dev_abs = np.std(diff_real_modelled_abs)    # Given as energy unit



        x_values = np.arange(0, len(y_real_demand), 1)

        labels = []
        for sorted_region in sorted_dict_real:
            if sorted_region in y_real_demand:
                geocode_lad = sorted_region[0]

                logging.debug(
                    "validation %s LAD %s: real: %s modelled: %s  modelled percentage: %s (%sp diff)",
                    fueltype_str,
                    geocode_lad,
                    round(result_dict['real_demand'][geocode_lad], 4),
                    round(result_dict['modelled_demand'][geocode_lad], 4),
                    round(100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad], 4),
                    round(100 - (100 / result_dict['real_demand'][geocode_lad] * result_dict['modelled_demand'][geocode_lad]), 4))

                labels.append(geocode_lad)

        # Calculate r_squared
        _slope, _intercept, r_value, _p_value, _std_err = stats.linregress(
            y_real_demand,
            y_modelled_demand)

        # --------
        # Axis
        # --------
        plt.tick_params(
            axis='x',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom='off',      # ticks along the bottom edge are off
            top='off',         # ticks along the top edge are off
            labelbottom='off') # labels along the bottom edge are off

        # ----------------------------------------------
        # Plot
        # ----------------------------------------------
        plt.plot(
            x_values,
            y_real_demand,
            linestyle='None',
            marker='o',
            alpha=0.7,
            markersize=1.6,
            fillstyle='full',
            markerfacecolor='grey',
            markeredgewidth=0.2,
            color=color_list[cnt_color],
            markeredgecolor=color_list[cnt_color],
            label='actual')

        plt.plot(
            x_values,
            y_modelled_demand,
            marker='o',
            linestyle='None',
            markersize=1.6,
            alpha=0.7,
            markerfacecolor='white',
            fillstyle='none',
            markeredgewidth=0.5,
            markeredgecolor=color_list[cnt_color],
            color=color_list[cnt_color],
            label='model' + label_list[cnt_color])

        # -----------
        # Labelling
        # -----------
        if label_points:
            for pos, txt in enumerate(labels):
                ax.text(
                    x_values[pos],
                    y_modelled_demand[pos],
                    txt,
                    horizontalalignment="right",
                    verticalalignment="top",
                    fontsize=1)

        font_additional_info = plotting_styles.font_info(size=3, color=color_list[cnt_color])

        title_info = ('R_2: {}, std_%: {} (GWh {}), av_diff_%: {} median_abs_dev: {}'.format(
            round(r_value, 2),
            round(std_dev_p, 2),
            round(std_dev_abs, 2),
            round(av_deviation_real_modelled, 2),
            round(median_absolute_deviation, 2)))

        plt.text(
            0.4,
            0.9 - cnt_color/10,
            title_info,
            ha='center',
            va='center',
            transform=ax.transAxes,
            fontdict=font_additional_info)

        cnt_color =+ 1

    plt.xlabel("UK regions (excluding northern ireland)")
    plt.ylabel("{} [GWh]".format(fueltype_str))
    plt.ylim(ymin=0)

    # --------
    # Legend
    # --------
    plt.legend(
        prop={
            'family': 'arial',
            'size': 6},
        frameon=False)

    # Tight layout
    plt.margins(x=0)
    plt.tight_layout()

    plt.savefig(fig_name)

    if plotshow:
        plt.show()
    else:
        plt.close()



"""Setting up the logger
"""
import os
import logging

def set_up_logger(path_log_file):
    """Create logger
    Arguments
    --------
    path_log_file : str
        Path to logger file
    Info
    -----
    The logging level can be changed depending on mode
    Note
    ----
    logger.debug('debug message')
    logger.warn('warn message')
    logger.error('error message')
    logger.critical('critical message')
    """
    # Create logging file if not existing
    if not os.path.isfile(path_log_file):
        open(path_log_file, 'w').close()

    # Set logger level
    logging.basicConfig(
        filename=path_log_file,
        filemode='w', #'a, w'
        level=logging.DEBUG, #INFO, DEBUG, ERROR, CRITICAL
        format=('%(asctime)s - %(name)s - %(levelname)s - %(message)s'))

    # Necessary to add loggers in visual studio console
    ##logging.getLogger().addHandler(logging.StreamHandler())

    # Turn on/off logger
    #logging.disable = False
    #logging.disable(logging.CRITICAL)



"""
"""
import os
from energy_demand.basic import lookup_tables
from energy_demand.assumptions import fuel_shares
from energy_demand.assumptions import general_assumptions
from energy_demand.assumptions import strategy_vars_def
from energy_demand.read_write import data_loader
from pkg_resources import resource_filename
from pkg_resources import Requirement
from pytest import mark

@mark.skip(reason="")
def test_assign_by_fuel_tech_p(config_file):
    """
    """
    config = data_loader.read_config_file(config_file)
    

    # Load data
    data = {}
    data['paths'] = config['CONFIG_DATA']
    data['lookups'] = lookup_tables.basic_lookups()
    data['enduses'], data['sectors'], data['fuels'], _, _ = data_loader.load_fuels(
        data['paths'])

    data['local_paths'] = data_loader.get_local_paths(path_main)

    #Load assumptions
    base_yr = 2015

    data['assumptions'] = general_assumptions.Assumptions(
        submodels_names=['a'],
        base_yr=base_yr,
        curr_yr=None,
        sim_yrs=None,
        paths=data['paths'],
        enduses=data['enduses'],
        sectors=data['sectors'],
        fueltypes=data['lookups']['fueltypes'],
        fueltypes_nr=data['lookups']['fueltypes_nr'])

    strategy_vars_def.load_param_assump(data['paths'], data['local_paths'], data['assumptions'])

    fuel_tech_p_by = fuel_shares.assign_by_fuel_tech_p( 
        data['enduses'],
        data['sectors'],
        data['lookups']['fueltypes'],
        data['lookups']['fueltypes_nr'])



"""
"""
import numpy as np
import matplotlib.pyplot as plt

from energy_demand import enduse_func
from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions

def run(results_every_year, lookups, path_plot_fig):
    """Plots

    Plot peak hour per fueltype over time for

    Arguments
    ---------
    tot_fuel_dh_peak : dict
        year, fueltype, peak_dh

    """
    # Set figure size
    fig = plt.figure(figsize=basic_plot_functions.cm2inch(14, 8))
    ax = fig.add_subplot(1, 1, 1)

    nr_y_to_plot = len(results_every_year)

    legend_entries = []

    # Initialise (number of enduses, number of hours to plot)
    y_init = np.zeros((lookups['fueltypes_nr'], nr_y_to_plot))

    for fueltype_str, fueltype in lookups['fueltypes'].items():
        fueltype_int = tech_related.get_fueltype_int(fueltype_str)

        # Legend
        legend_entries.append(fueltype_str)

        # Read out fueltype specific load
        data_over_years = []
        for model_year_object in results_every_year.values():

            # Sum fuel across all regions
            fuel_all_regs = np.sum(model_year_object, axis=1) # (fueltypes, 8760 hours)

            # Get peak day across all enduses for every region
            _, gw_peak_fueltyp_h = enduse_func.get_peak_day_single_fueltype(fuel_all_regs[fueltype_int])

            # Add peak hour
            data_over_years.append(gw_peak_fueltyp_h)

        y_init[fueltype] = data_over_years

    # ----------
    # Plot lines
    # ----------
    #linestyles = plotting_styles.linestyles()

    years = list(results_every_year.keys())
    for fueltype, _ in enumerate(y_init):
        plt.plot(
            years,
            y_init[fueltype],
            linewidth=0.7)

    ax.legend(
        legend_entries,
        prop={
            'family': 'arial',
            'size': 8},
        frameon=False)

    # -
    # Axis
    # -
    base_yr = 2015
    major_interval = 10
    minor_interval = 5

    # Major ticks
    major_ticks = np.arange(base_yr, years[-1] + major_interval, major_interval)
    ax.set_xticks(major_ticks)

    # Minor ticks
    minor_ticks = np.arange(base_yr, years[-1] + minor_interval, minor_interval)
    ax.set_xticks(minor_ticks, minor=True)

    plt.xlim(2015, years[-1])

    # --------
    # Labeling
    # --------
    plt.ylabel("GW")
    plt.xlabel("year")
    plt.title("ED peak hour, y, all enduses and regions")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    # Save fig
    fig.savefig(path_plot_fig)
    plt.close()



"""All plots of paper II
"""
import os
from collections import defaultdict
import numpy as np

from energy_demand.read_write import data_loader, read_data
from energy_demand.basic import date_prop
from energy_demand.basic import lookup_tables
from energy_demand.plotting import fig_weather_variability_priod
from energy_demand.plotting import fig_total_demand_peak
from energy_demand.plotting import fig_p2_temporal_validation
from energy_demand.validation import elec_national_data
from energy_demand.technologies import tech_related
from energy_demand.plotting import fig_p2_annual_hours_sorted
from energy_demand.plotting import fig_p2_spatial_val
from energy_demand import enduse_func

def main(
        path_data_ed,
        path_shapefile_input,
        path_out_plots,
        plot_crit_dict):
    """Figure II plots
    """

    # ---------------------------------------------------------
    # Iterate folders and read out all weather years and stations
    # ---------------------------------------------------------
    all_result_folders = os.listdir(path_data_ed)
    paths_folders_result = []
    weather_yrs = []
    weather_station_per_y = {}
    all_calculated_yrs_paths = []
    for result_folder in all_result_folders:
        try:
            split_path_name = result_folder.split("__")
            weather_yr = int(split_path_name[0])
            weather_yrs.append(weather_yr)
            try:
                weather_station = int(split_path_name[1])
            except:
                weather_station = "all_stations"
            try:
                weather_station_per_y[weather_yr].append(weather_station)
            except:
                weather_station_per_y[weather_yr] = [weather_station]

            # Collect all paths to simulation result folders
            paths_folders_result.append(
                os.path.join(path_data_ed, result_folder))
            
            tupyle_yr_path = (weather_yr, os.path.join(path_data_ed, result_folder))
            all_calculated_yrs_paths.append(tupyle_yr_path)

        except ValueError:
            pass

    # -----------
    # Used across different plots
    # -----------
    data = {}
    data['lookups'] = lookup_tables.basic_lookups()

    data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(
        os.path.join(path_data_ed))

    data['assumptions']['seasons'] = date_prop.get_season(year_to_model=2015)
    data['assumptions']['model_yeardays_daytype'], data['assumptions']['yeardays_month'], data['assumptions']['yeardays_month_days'] = date_prop.get_yeardays_daytype(year_to_model=2015)

    population_data = read_data.read_scenaric_population_data(
        os.path.join(path_data_ed, 'model_run_pop'))

    ####################################################################
    # Plotting weather variability results for all weather stations (Fig 2b)
    ####################################################################
    weather_yr_container = defaultdict(dict)

    for weather_yr, result_folder in all_calculated_yrs_paths:
        results_container = read_data.read_in_results(
            os.path.join(result_folder, 'model_run_results_txt'),
            data['assumptions']['seasons'],
            data['assumptions']['model_yeardays_daytype'])

        weather_yr_container['tot_fueltype_yh'][weather_yr] = results_container['tot_fueltype_yh'] #tot_fueltype_yh
        weather_yr_container['results_enduse_every_year'][weather_yr] = results_container['ed_fueltype_regs_yh']

    # ####################################################################
    # Create plot with regional and non-regional plots for second paper
    # Compare hdd calculations and disaggregation of regional and local
    # ####################################################################
    if plot_crit_dict['plot_scenarios_sorted']:

        fig_p2_annual_hours_sorted.run(
            data_input=weather_yr_container['results_enduse_every_year'],
            regions=data['regions'],
            simulation_yrs_to_plot=[2015], # Simulation year to plot
            fueltype_str='electricity',
            path_shapefile=path_shapefile_input,
            fig_name=os.path.join(path_out_plots, "fig_paper_IIb_weather_var_map.pdf"))

def plot_fig_spatio_temporal_validation(
        path_regional_calculations,
        path_rolling_elec_demand,
        path_temporal_elec_validation,
        path_temporal_gas_validation,
        path_non_regional_elec_2015,
        path_out_plots,
        plot_show=False
    ):
    """
    Create plot with regional and non-regional plots for second paper
    Compare hdd calculations and disaggregation of regional and local
    """
    # ---------------------------------------------------------
    # Iterate folders and read out all weather years and stations
    # ---------------------------------------------------------
    all_result_folders = os.listdir(path_regional_calculations)

    paths_folders_result = []
    data_container = defaultdict(dict)
    ed_fueltype_regs_yh = defaultdict(dict)
    weather_yr_station_tot_fueltype_yh = defaultdict(dict)
    residential_results = defaultdict(dict)
    for scenario_folder in all_result_folders:
        result_folders = os.listdir(os.path.join(path_regional_calculations, scenario_folder))
        for result_folder in result_folders:
            try:
                split_path_name = result_folder.split("__")
                weather_yr = int(split_path_name[0])

                try:
                    weather_station = int(split_path_name[1])
                except:
                    weather_station = "all_stations"

                paths_folders_result.append(
                    os.path.join(path_regional_calculations, result_folder))

                data = {}
                data['lookups'] = lookup_tables.basic_lookups()
                data['enduses'], data['assumptions'], data['regions'] = data_loader.load_ini_param(
                    os.path.join(path_regional_calculations, all_result_folders[0])) # last result folder
                data['assumptions']['seasons'] = date_prop.get_season(year_to_model=2015)
                data['assumptions']['model_yeardays_daytype'], data['assumptions']['yeardays_month'], data['assumptions']['yeardays_month_days'] = date_prop.get_yeardays_daytype(year_to_model=2015)

                results_container = read_data.read_in_results(
                    os.path.join(
                        path_regional_calculations,
                        scenario_folder,
                        "{}__{}".format(weather_yr, weather_station),
                        'model_run_results_txt'),
                    data['assumptions']['seasons'],
                    data['assumptions']['model_yeardays_daytype'])

                weather_yr_station_tot_fueltype_yh[weather_yr][weather_station] =  results_container['tot_fueltype_yh']
                ed_fueltype_regs_yh[weather_yr][weather_station] = results_container['ed_fueltype_regs_yh']
                residential_results[weather_yr][weather_station] = results_container['residential_results']
            except ValueError:
                pass

    data_container['ed_fueltype_regs_yh'] = ed_fueltype_regs_yh
    data_container['tot_fueltype_yh'] = weather_yr_station_tot_fueltype_yh
    data_container['residential_results'] = residential_results
    data_container = dict(data_container)

    # -------------------------------------------------
    # Collect non regional 2015 elec data
    # Calculated with all regional weather stations
    # -------------------------------------------------
    year_non_regional = 2015
    path_with_txt = os.path.join(
        path_non_regional_elec_2015,
        "{}__{}".format(str(year_non_regional), "all_stations"),
        'model_run_results_txt')

    demand_year_non_regional = read_data.read_in_results(
        path_with_txt,
        data['assumptions']['seasons'],
        data['assumptions']['model_yeardays_daytype'])
    tot_fueltype_yh = demand_year_non_regional['tot_fueltype_yh']

    fueltype_int = tech_related.get_fueltype_int('electricity')

    non_regional_elec_2015 = tot_fueltype_yh[year_non_regional][fueltype_int]

    # ---Collect real electricity data of year 2015
    elec_2015_indo, _ = elec_national_data.read_raw_elec_2015(
        path_rolling_elec_demand)

    # Factor data as total sum is not identical
    f_diff_elec = np.sum(non_regional_elec_2015) / np.sum(elec_2015_indo)
    elec_factored_yh = f_diff_elec * elec_2015_indo

    # *****************************************************************
    # Temporal validation
    # Compare regional and non regional and actual demand over time
    # *****************************************************************
    simulation_yr_to_plot = 2015

    winter_week, spring_week, summer_week, autumn_week = date_prop.get_seasonal_weeks()

    # Peak day
    peak_day, _ = enduse_func.get_peak_day_single_fueltype(elec_factored_yh)

    # Convert days to hours
    period_to_plot = list(range(0, 400))
    period_to_plot = date_prop.get_8760_hrs_from_yeardays(winter_week)
    period_to_plot = date_prop.get_8760_hrs_from_yeardays([peak_day])
    period_to_plot_winter = date_prop.get_8760_hrs_from_yeardays(winter_week)
    period_to_plot_spring = date_prop.get_8760_hrs_from_yeardays(spring_week)

    fig_p2_temporal_validation.run_fig_p2_temporal_validation(
        data_input=data_container['tot_fueltype_yh'],
        weather_yr=2015,
        fueltype_str='electricity',
        simulation_yr_to_plot=simulation_yr_to_plot, # Simulation year to plot
        period_h=period_to_plot,
        validation_elec_2015=elec_factored_yh,
        non_regional_elec_2015=non_regional_elec_2015,
        fig_name=os.path.join(path_out_plots, "temporal_validation_elec.pdf"),
        titel="yearday: {}".format(peak_day),
        y_lim_val=55,
        plot_validation=False,
        plot_show=plot_show)
    
    fueltype_gas = tech_related.get_fueltype_int('gas')
    fig_p2_temporal_validation.run_fig_p2_temporal_validation(
        data_input=data_container['tot_fueltype_yh'],
        weather_yr=2015,
        fueltype_str='gas',
        simulation_yr_to_plot=simulation_yr_to_plot, # Simulation year to plot
        period_h=period_to_plot,
        validation_elec_2015=None,
        non_regional_elec_2015=tot_fueltype_yh[year_non_regional][fueltype_gas],
        fig_name=os.path.join(path_out_plots, "temporal_validation_gas.pdf"),
        titel="yearday: {}".format(peak_day),
        y_lim_val=250,
        plot_validation=False,
        plot_show=plot_show)

    # -------------------
    # Spatial validation (not with maps)
    # -------------------
    # non_regional: All weather station, spatially disaggregated TODO Give BETTER NAMES
    # regional: Only one weather station for whole countr but still data for every region
    weather_yr = 2015
    fig_p2_spatial_val.run(
        simulation_yr_to_plot=simulation_yr_to_plot,
        demand_year_non_regional=demand_year_non_regional['residential_results'][weather_yr],
        demand_year_regional=data_container['residential_results'][weather_yr],
        fueltypes=data['lookups']['fueltypes'],
        fig_path=path_out_plots,
        path_temporal_elec_validation=path_temporal_elec_validation,
        path_temporal_gas_validation=path_temporal_gas_validation,
        regions=data['regions'],
        plot_crit=plot_show)

    # -------------------
    # Spatial validation (plot regional maps)
    # with deviation from average estimated demand across scenarios
    # -------------------



"""This file creates dummy data needed specifically for the energy_demand model
"""
import os
import numpy as np
from pkg_resources import Requirement
from pkg_resources import resource_filename
import configparser

from energy_demand.read_write import write_data
from energy_demand.read_write import data_loader
from energy_demand.basic import basic_functions
from energy_demand.scripts.s_rs_raw_shapes import run
from energy_demand.assumptions import general_assumptions
from energy_demand.basic import lookup_tables

def create_folders_to_file(path_to_file, attr_split):
    """
    """
    path = os.path.normpath(path_to_file)

    path_up_to_raw_folder = path.split(attr_split)[0]
    path_after_raw_folder = path.split(attr_split)[1]

    folders_to_create = path_after_raw_folder.split(os.sep)

    path_curr_folder = os.path.join(path_up_to_raw_folder, attr_split)

    for folder in folders_to_create[1:-1]: #Omit first entry and file
        path_curr_folder = os.path.join(path_curr_folder, folder)
        basic_functions.create_folder(path_curr_folder)

def dummy_sectoral_load_profiles(local_paths, path_main):
    """Create dummy sectoral load profiles

    Arguments
    ---------
    local_paths : dict
        Paths
    path_main : str
        Main path
    """
    create_folders_to_file(
        os.path.join(local_paths['ss_load_profile_txt'], "dumm"), "_processed_data")

    paths = data_loader.load_paths(path_main)

    dict_enduses, dict_sectors, _, _, _ = data_loader.load_fuels(paths)

    for enduse in dict_enduses['service']:
        for sector in dict_sectors['service']:

            joint_string_name = str(sector) + "__" + str(enduse)

            # Flat profiles
            load_peak_shape_dh = np.full((24), 1)
            shape_non_peak_y_dh = np.full((365, 24), 1/24)
            shape_non_peak_yd = np.full((365), 1/365)

            write_data.create_txt_shapes(
                joint_string_name,
                local_paths['ss_load_profile_txt'],
                load_peak_shape_dh,
                shape_non_peak_y_dh,
                shape_non_peak_yd)

def post_install_setup_minimum(args):
    """If not all data are available, this scripts allows to
    create dummy datas (temperature and service sector load profiles)

    """
    path_config_file = args.config_file

    config = data_loader.read_config_file(path_config_file)
    path_local_data = config['PATHS']['path_local_data']

    # ==========================================
    # Post installation setup witout access to non publicy available data
    # ==========================================
    print("... running initialisation scripts with only publicly available data")

    local_paths = data_loader.get_local_paths(path_config_file)

    # Create folders to input data
    raw_folder = os.path.join(path_local_data, '_raw_data')

    basic_functions.create_folder(raw_folder)
    basic_functions.create_folder(config['PATHS']['path_processed_data'])
    basic_functions.create_folder(local_paths['path_post_installation_data'])
    basic_functions.create_folder(local_paths['load_profiles'])
    basic_functions.create_folder(local_paths['rs_load_profile_txt'])
    basic_functions.create_folder(local_paths['ss_load_profile_txt'])

    # Load data
    data = {}

    data['paths'] = data_loader.load_paths(path_config_file)

    data['lookups'] = lookup_tables.basic_lookups()

    data['enduses'], data['sectors'], data['fuels'], lookup_enduses, lookup_sector_enduses = data_loader.load_fuels(data['paths'])

    # Assumptions
    data['assumptions'] = general_assumptions.Assumptions(
        lookup_enduses=lookup_enduses,
        lookup_sector_enduses=lookup_sector_enduses,
        base_yr=2015,
        weather_by=config['CONFIG']['user_defined_weather_by'],
        simulation_end_yr=config['CONFIG']['user_defined_simulation_end_yr'],
        paths=data['paths'],
        enduses=data['enduses'],
        sectors=data['sectors'])

    # Read in residential submodel shapes
    run(data['paths'], local_paths, config['CONFIG']['base_yr'])

    # --------
    # Dummy service sector load profiles
    # --------
    dummy_sectoral_load_profiles(
        local_paths, path_config_file)

    print("Successfully finished post installation setup with open source data")



from ruamel.yaml import YAML

from energy_demand.read_write import write_data


# ------------------------
# TEmp data
# ------------------------
path = "C:/_scrap/temp-data.yml"

variant_list = []

for i in range(1, 101, 1):

    name = "NF{}".format(i)

    entry = {
        'name': name,
        'description': 'Weather realisation {}'.format(name),
        'data': {
            't_min': 'weather_data_{}.csv'.format(name),
            't_max': 'weather_data_{}.csv'.format(name)}
        }
    variant_list.append(entry)

dict_out = {
    'variants': variant_list
}

write_data.write_yaml(dict_out, path)

# ------------------------
# Weather Stations
# ------------------------
path = "C:/_scrap/weateher-station.yml"

variant_list = []

for i in range(1, 101, 1):

    name = "NF{}".format(i)

    entry = {
        'name': "weather_station_coordinates__{}".format(name),
        'description': 'Weather station coordinates for weather data {}'.format(name),
        'data': {
            'longitude': 'stations_{}.csv'.format(name),
            'latitude': 'stations_{}.csv'.format(name)}
        }
    variant_list.append(entry)

dict_out = {
    'variants': variant_list
}

write_data.write_yaml(dict_out, path)



"""Functions to configure energy demand outputs for supply model
"""
import os
import logging
import numpy as np
import pandas as pd

from energy_demand.basic import date_prop, testing_functions, lookup_tables

def constrained_results(
        results_constrained,
        #results_unconstrained,
        results_unconstrained_no_heating,
        submodels_names,
        technologies,
        reg_nrs,
        fueltype_nrs
    ):
    """Prepare results for energy supply model for
    constrained model running mode (no heat is provided but
    technology specific fuel use).
    The results for the supply model are provided aggregated
    as follows:

        { "submodel_fueltype_tech": np.array(regions, timesteps)}

    Because SMIF only takes results in the
    form of {key: np.array(regions, timesteps)}, the key
    needs to contain information about submodel, fueltype,
    and technology. Also these key must be defined in
    the `submodel_model` configuration file.

    Arguments
    ----------
    results_constrained : dict
        Aggregated results in form
        {technology: np.array((sector, region, fueltype, timestep))}
    results_unconstrained : array
        Restuls of unconstrained mode
        np.array((sector, regions, fueltype, timestep))
    submodels_names : list
        Names of sectors fur supply model
    technologies : dict
        Technologies

    Returns
    -------
    supply_results : dict
        No technology specific delivery (heat is provided in form of a fueltype)
        {submodel_fueltype: np.array((region, intervals))}

    Note
    -----
    For the fuel demand for CHP plants, the co-generated electricity
    is not included in the demand model. Additional electricity supply
    generated from CHP plants need to be calculated in the supply
    model based on the fuel demand for CHP.
    For CHP efficiency therefore not the overall efficiency is used
    but only the thermal efficiency
    """
    supply_results = {}
    fueltypes = lookup_tables.basic_lookups()['fueltypes']

    # ----------------------------------------
    # Add all constrained results (technology specific results)
    # Aggregate according to submodel, fueltype, technology, region, timestep
    # ----------------------------------------
    #summed_heating_demand = np.zeros((len(submodels_names), reg_nrs, fueltype_nrs, 8760))
    for submodel_nr, submodel in enumerate(submodels_names):
        for tech, fuel_tech in results_constrained.items():
            # ----
            # Technological simplifications because of different technology definition and because not all technologies are used in supply model
            # ----
            tech_simplified = model_tech_simplification(tech)
            fueltype_str = technologies[tech_simplified].fueltype_str
            fueltype_int = technologies[tech_simplified].fueltype_int
            key_name = "{}_{}_{}".format(submodel, fueltype_str, tech_simplified)

            # Out results
            supply_results[key_name] = fuel_tech[submodel_nr][:, fueltype_int, :]

    #for key_name, values in supply_results.items():
    #    if testing_functions.test_if_minus_value_in_array(values):
    #        raise Exception("TTT {} {}".format(key_name, np.sum(values)))
    #--------------------------------
    # Get all non heating related enduse
    # --------------------------------
    #assert not testing_functions.test_if_minus_value_in_array(results_unconstrained)
    #assert not testing_functions.test_if_minus_value_in_array(summed_heating_demand)
    #print(fuel_tech[submodel_nr][:, fueltype_int, :].shape)
    #print("ddddd")
    #print(np.sum(summed_heating_demand))
    #print(np.sum(results_unconstrained))
    #print(np.sum(results_unconstrained_no_heating))
    #print(results_unconstrained.shape)
    #print(summed_heating_demand.shape)
    #print(results_unconstrained_no_heating.shape)
    #print("DDdf --")
    #assert not testing_functions.test_if_minus_value_in_array(results_unconstrained_no_heating)
    # Substract constrained fuel from nonconstrained (total) fuel
    #non_heating_ed_OLD = results_unconstrained - summed_heating_demand #sum(results_constrained.values())

    #print(np.sum(non_heating_ed_OLD))
    #print(np.sum(non_heating_ed))
    #print(non_heating_ed_OLD.shape)
    #print(non_heating_ed.shape)
    assert not testing_functions.test_if_minus_value_in_array(results_unconstrained_no_heating)

    # ---------------------------------
    # Add non_heating for all fueltypes
    # ---------------------------------
    for submodel_nr, submodel in enumerate(submodels_names):
        for fueltype_str, fueltype_int in fueltypes.items():

            if fueltype_str == 'heat':
                pass #Do not add non_heating demand for fueltype heat
            else:
                key_name = "{}_{}_{}".format(submodel, fueltype_str, "non_heating")

                # Add fuel for all regions for specific fueltype
                supply_results[key_name] = results_unconstrained_no_heating[submodel_nr][:, fueltype_int, :]

    # --------------------------------------------
    # Check whether any entry is smaller than zero
    # --------------------------------------------
    for key_name, values in supply_results.items():
        if testing_functions.test_if_minus_value_in_array(values):
            logging.info("info tt: {}  {}".format(values, np.sum(values)))
            raise Exception("Error d: Negative entry in results {} {}".format(key_name, np.sum(values)))

    return supply_results

def unconstrained_results(
        results_unconstrained,
        submodels_names
    ):
    """Prepare results for energy supply model for
    unconstrained model running mode (heat is provided).
    The results for the supply model are provided aggregated
    for every submodel, fueltype, region, timestep

    Note
    -----
    Because SMIF only takes results in the
    form of {key: np.aray(regions, timesteps)}, the key
    needs to contain information about submodel and fueltype

    Also these key must be defined in the `submodel_model`
    configuration file

    Arguments
    ----------
    results_unconstrained : array
        Results of unconstrained mode
        np.array((sector, regions, fueltype, timestep))
    submodels_names : list
        Names of sectors for supply model

    Returns
    -------
    supply_results : dict
        No technology specific delivery (heat is provided in form of a fueltype)
        {submodel_fueltype: np.array((region, intervals))}
    """
    supply_results = {}
    fueltypes = lookup_tables.basic_lookups()['fueltypes']

    for submodel_nr, submodel in enumerate(submodels_names):
        for fueltype_str, fueltype_int in fueltypes.items():

            # Generate key name (must be defined in `sector_models`)
            key_name = "{}_{}".format(submodel, fueltype_str)

            # Add fueltype specific demand for all regions
            supply_results[key_name] = results_unconstrained[submodel_nr][:, fueltype_int, :]

            assert not testing_functions.test_if_minus_value_in_array(supply_results[key_name])

    logging.info("... Prepared results for energy supply model in unconstrained mode")
    return supply_results

def model_tech_simplification(tech):
    """This function aggregated different technologies
    which are not defined in supply model

    Arguments
    ---------
    tech : str
        Technology

    Returns
    -------
    tech_newly_assigned : str
        Technology newly assigned
    """
    if tech == 'boiler_condensing_gas':
        tech_newly_assigned = 'boiler_gas'
    elif tech == 'boiler_condensing_oil':
        tech_newly_assigned = 'boiler_oil'
    elif tech == 'storage_heater_electricity':
        tech_newly_assigned = 'boiler_electricity'
    elif tech == 'secondary_heater_electricity':
        tech_newly_assigned = 'boiler_electricity'
    else:
        tech_newly_assigned = tech

    if tech != tech_newly_assigned:
        logging.debug(
            "Assigned new technology '%s' for '%s' to provide for simplified supply output",
            tech_newly_assigned, tech)

    return tech_newly_assigned

def write_national_results_amman(
        path_folder,
        results_unconstrained,
        regions,
        fueltype_str,
        fuelype_nr,
        year
    ):
    """Write national results of all fueltypes

    Inputs
    ------
    results_unconstrained : array
        (submodel, region, fueltype, periods)
    """
    logging.info("... writing file for amman")

    path = os.path.join(path_folder, "file_AMMAN_{}_{}.csv".format(fueltype_str, year))

    # Sum across all sectors
    sum_across_submodels = results_unconstrained.sum(axis=0)

    # Change ot days
    nr_regs = len(regions)
    nr_fueltypes = sum_across_submodels.shape[1]
    sum_across_regions_days = sum_across_submodels.reshape(nr_regs, nr_fueltypes, 365, 24)

    # Iterate over every hour in year
    rows = []
    for region_nr, region in enumerate(regions):
        for day in range(365):
            row = {'region': region}

            daysum = np.sum(sum_across_regions_days[region_nr][fuelype_nr][day])
            row['day'] = day
            row['demand_GWh'] = daysum

            rows.append(row)

    # Create dataframe
    col_names = ['region', 'day', 'demand_GWh']
    my_df = pd.DataFrame(rows, columns=col_names)

    my_df.to_csv(path, index=False) #Index prevents writing index rows

def write_national_results(
        path_folder,
        results_unconstrained,
        enduse_specific_results,
        fueltype_str,
        fuelype_nr,
        year,
        write_regions=False,
        submodels_names=['residential', 'service', 'industry']
    ):
    """Write national results of all fueltypes

    Input
    ------
    results_unconstrained : array
        (submodel, region, fueltype, periods)
    """
    logging.info("... writing file for modassar")

    path = os.path.join(path_folder, "file_MODASSAR_{}_{}.csv".format(fueltype_str, year))

    # Sum across all regions
    sum_across_regions = results_unconstrained.sum(axis=1)

    rows = []
    for hour in range(8760):

        # Get day and hour
        day_year, hour_day_year = date_prop.convert_h_to_day_year_and_h(hour)

        row = {'year': year, 'hour': hour}

        for submodel_nr, submodel in enumerate(submodels_names):

            # Total energy demand
            ed_submodel_h = sum_across_regions[submodel_nr][fuelype_nr][hour]

            # Space heating related demand for sector
            if submodel_nr == 0:
                space_heating_demand = enduse_specific_results['rs_space_heating'][fuelype_nr][day_year][hour_day_year]
            elif submodel_nr == 1:
                space_heating_demand = enduse_specific_results['ss_space_heating'][fuelype_nr][day_year][hour_day_year]
            else:
                space_heating_demand = enduse_specific_results['is_space_heating'][fuelype_nr][day_year][hour_day_year]

            ed_submodel_heating_h = space_heating_demand
            str_name_heat = "{}_heat".format(submodel)
            row[str_name_heat] = ed_submodel_heating_h

            # Non-heating related demand
            ed_submodel_non_heating_h = ed_submodel_h - space_heating_demand

            str_name_non_heat = "{}_non_heat".format(submodel)

            row[str_name_non_heat] = ed_submodel_non_heating_h

        rows.append(row)

    # Create dataframe
    col_names = [
        'year',
        'hour',
        'residential_non_heat',
        'residential_heat',
        'service_non_heat',
        'service_heat',
        'industry_non_heat',
        'industry_heat']

    my_df = pd.DataFrame(rows, columns=col_names)
    my_df.to_csv(path, index=False) #Index prevents writing index rows

    # Plot national results
    '''validation_enduses.plot_dataframe_function(
        my_df,
        x_column_name='hour',
        y_column_names=['residential_non_heat', 'service_non_heat', 'industry_non_heat'],
        plot_kind='line')'''




"""
"""
import os
from collections import defaultdict
import numpy as np
import pandas as pd
import geopandas as gpd
from shapely.geometry import Point
import matplotlib.pyplot as plt

from energy_demand.plotting import result_mapping
from energy_demand.plotting import fig_p2_weather_val
from energy_demand.read_write import data_loader
from energy_demand.read_write import read_weather_data
from energy_demand.plotting import basic_plot_functions

def run(
        path_to_weather_data,
        folder_path_weater_stations,
        path_shapefile,
        fig_path
    ):
    """

    """
    # ----------------------------------------------------------
    # Iterate all calculated weather stations
    # ----------------------------------------------------------
    weather_yrs_stations = read_weather_data.get_all_station_per_weather_yr(path_to_weather_data)

    weather_station_coordinates = data_loader.read_weather_stations_raw(folder_path_weater_stations)
    # ----------------------------------------------------------
    # Station nr and how many times over simulation period
    # ----------------------------------------------------------
    station_counting_over_yr = {}

    for weather_yr, stations in weather_yrs_stations.items():
        for station in stations:
            try:
                station_counting_over_yr[station].append(weather_yr)
            except:
                station_counting_over_yr[station] = [weather_yr]

    # Count number of years
    for station in station_counting_over_yr:
        station_counting_over_yr[station] = len(station_counting_over_yr[station])

    # ----------------------------------------------------------
    # Create dataframe
    #{station_id: all_weather_stations[weather_yr][station_id]}
    #df = pd.DataFrame({'src_id': [...], 'longitude': [...], 'latitude': [...]})
    #  ----------------------------------------------------------
    stations_as_dict = {}
    for station_id in station_counting_over_yr:
        try:
            stations_as_dict[station_id] = {
                'longitude': weather_station_coordinates[station_id]['longitude'],
                'latitude': weather_station_coordinates[station_id]['latitude'],
                'nr_of_weather_yrs': station_counting_over_yr[station_id]}
        except:
            print("The station nr {} coul not be found".format(station_id))

    df = pd.DataFrame.from_dict(stations_as_dict, orient='index')

    df['Coordinates'] = list(zip(df.longitude, df.latitude))
    df['Coordinates'] = df['Coordinates'].apply(Point)

    # Load uk shapefile
    uk_shapefile = gpd.read_file(path_shapefile)

    # Assign correct projection
    crs = {'init': 'epsg:27700'} #27700 == OSGB_1936_British_National_Grid
    uk_gdf = gpd.GeoDataFrame(uk_shapefile, crs=crs)

    # Plot
    ax = uk_gdf.plot(
        linewidth=0.3,
        color='whitesmoke',
        edgecolor='black',
        figsize=basic_plot_functions.cm2inch(25, 20))
        
    # plot coordinates
    crs = {'init': 'epsg:4326'}
    gdf = gpd.GeoDataFrame(df, geometry='Coordinates', crs=crs)
    gdf = gdf.to_crs({'init' :'epsg:27700'})

    field_to_plot = 'nr_of_weather_yrs'
    nr_of_intervals = 6

    bin_values = result_mapping.get_reasonable_bin_values_II(
        data_to_plot=list(gdf[field_to_plot]),
        nr_of_intervals=nr_of_intervals)

    print("BINS " + str(bin_values))
    gdf, cmap_rgb_colors, color_zero, min_value, max_value = fig_p2_weather_val.user_defined_bin_classification(
        gdf,
        field_to_plot,
        bin_values=bin_values,
        cmap_sequential='PuRd') #'Reds')

    # plot with face color attribute
    gdf.plot(
        ax=ax,
        markersize=20,
        facecolor=gdf['bin_color'],
        edgecolor='black',
        linewidth=0.3)

    legend_handles = result_mapping.add_simple_legend(
        bin_values,
        cmap_rgb_colors,
        color_zero,
        patch_form='circle')

    plt.legend(
        handles=legend_handles,
        title=str(field_to_plot),
        prop={'size': 8},
        loc='center left', bbox_to_anchor=(1, 0.5),
        frameon=False)
    try:
        os.remove(fig_path)
    except:
        pass
    plt.savefig(fig_path)

#def plot_map_


"""This file contains all functions to generate a flat load profile
"""
import numpy as np

def flat_shape(nr_of_days=365):
    """Create completely flat shape for peak and non-peak

    Arguments
    ---------
    nr_of_days : int
        Nr of modelled days

    Returns
    -------
    shape_peak_dh : array
        Dh shape for peak day
    flat_shape_y_dh : array
        Shape non peak dh
    flat_shape_yd : array
        Shape yd for non peak
    """
    flat_shape_yd = np.ones((nr_of_days), dtype="float") / nr_of_days

    flat_shape_yh = np.full((nr_of_days, 24), 1/(nr_of_days * 24), dtype="float")

    # Flat shape, ever hour same amount
    flat_shape_y_dh = np.full((nr_of_days, 24), (1.0/24), dtype="float")

    return flat_shape_yd, flat_shape_yh, flat_shape_y_dh

class GenericFlatEnduse(object):
    """Class for creating generic enduses with flat shapes,
    i.e. same amount of fuel for every hour in a year.

    Arguments
    ---------
    enduse_fuel : float
        Yearly total fuel
    """
    def __init__(self, enduse_fuel):

        _, flat_shape_yh, _ = flat_shape()

        # Yh fuel shape per fueltype (non-peak)
        self.fuel_yh = enduse_fuel[:, np.newaxis, np.newaxis] * flat_shape_yh[np.newaxis, :, :]

        self.flat_profile_crit = True



import os
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd
import geopandas as gpd

from energy_demand.read_write import data_loader, read_data
from energy_demand.basic import date_prop
from energy_demand.basic import basic_functions
from energy_demand.basic import lookup_tables
from energy_demand.technologies import tech_related
from energy_demand.plotting import basic_plot_functions
from energy_demand.plotting import result_mapping
from energy_demand.plotting import fig_p2_weather_val

def total_demand_national_scenarios(
        scenario_result_paths,
        sim_yrs,
        fueltype_str,
        path_out_plots
    ):
    dict_scenarios_weather_yrs = {}

    columns = [
        'weather_yr',
        'national_peak_demand']

    fueltype_int = tech_related.get_fueltype_int(fueltype_str)

    # ----------------
    # Read all data inform of scenario, simulationyr, weather_yrs
    # ----------------
    for scenario_path in scenario_result_paths:
        scenario_name = os.path.split(scenario_path)[-1]
        dict_scenarios_weather_yrs[scenario_name] = {}

        weather_yrs = []
        # Get all folders with weather_yr run results (name of folder is scenario)
        weather_yr_scenarios_paths = os.listdir(scenario_path)

        for simulation_run in weather_yr_scenarios_paths:   
            if simulation_run != '_results_PDF_figs':

                weather_yr_scenarios_paths = os.listdir(os.path.join(scenario_path, simulation_run))     

                for weather_yr_scenario_path in weather_yr_scenarios_paths:
                    try:
                        split_path_name = weather_yr_scenario_path.split("__")
                        weather_yr = int(split_path_name[0])
                        path_to_weather_yr = os.path.join(scenario_path, simulation_run, "{}__{}".format(weather_yr, 'all_stations'))
                        weather_yrs.append((weather_yr, path_to_weather_yr))
                    except:
                        pass

        for simulation_yr in sim_yrs:
            dict_scenarios_weather_yrs[scenario_name][simulation_yr] = pd.DataFrame(columns=columns)
            for weather_yr, path_to_weather_yr in weather_yrs:
            
                seasons = date_prop.get_season(year_to_model=2015)
                model_yeardays_daytype, _, _ = date_prop.get_yeardays_daytype(year_to_model=2015)

                results_container = read_data.read_in_results(
                    os.path.join(path_to_weather_yr, 'model_run_results_txt'),
                    seasons,
                    model_yeardays_daytype)

                # ---------------------------------------------------
                # Calculate hour with national peak demand
                # This may be different depending on the weather yr
                # ---------------------------------------------------
                ele_regions_8760 = results_container['ed_fueltype_regs_yh'][simulation_yr][fueltype_int]
                sum_all_regs_fueltype_8760 = np.sum(ele_regions_8760, axis=0) # Sum for every hour

                max_day = int(basic_functions.round_down((np.argmax(sum_all_regs_fueltype_8760) / 24), 1))
                max_h = np.argmax(sum_all_regs_fueltype_8760)
                max_demand = np.max(sum_all_regs_fueltype_8760)

                # Calculate the national peak demand in GW
                national_peak_GW = np.max(sum_all_regs_fueltype_8760)

                # -----------------------
                # Add to final container
                # -----------------------
                line_entry = [[
                    weather_yr,
                    national_peak_GW
                    ]]

                line_df = pd.DataFrame(line_entry, columns=columns)
                existing_df = dict_scenarios_weather_yrs[scenario_name][simulation_yr] 
                appended_df = existing_df.append(line_df)

                dict_scenarios_weather_yrs[scenario_name][simulation_yr] = appended_df


    # ------------------------------------------------------------------------------------------
    # Create plot
    # ------------------------------------------------------------------------------------------
    print("....create plot")

    weather_yr_to_plot = 1979 #TODO

    color_list = ['red', 'green', 'orange', '#37AB65', '#C0E4FF', '#3DF735', '#AD6D70', '#EC2504', '#8C0B90', '#27B502', '#7C60A8', '#CF95D7', '#F6CC1D']
    
    # Calculate quantiles
    quantile_95 = 0.95
    quantile_05 = 0.05

    # Create dataframe with rows as scenario and lines as simulation yrs
    scenarios = list(dict_scenarios_weather_yrs.keys())

    # Containers
    df_total_demand_2015 = pd.DataFrame(columns=scenarios)
    df_q_95_scenarios = pd.DataFrame(columns=scenarios)
    df_q_05_scenarios = pd.DataFrame(columns=scenarios)

    for simulation_yr in sim_yrs:

        line_entries_95 = []
        line_entries_05 = []
        line_entries_tot_h = []

        for scenario_name in scenarios:
            print("-- {}  {}".format(scenario_name, simulation_yr))
            # Calculate entires over year
            df_weather_yrs = dict_scenarios_weather_yrs[scenario_name][simulation_yr]

            df_q_95 = df_weather_yrs['national_peak_demand'].quantile(quantile_95)
            df_q_05 = df_weather_yrs['national_peak_demand'].quantile(quantile_05)

            peak_weather_yr_2015 = df_weather_yrs[df_weather_yrs['weather_yr']==weather_yr_to_plot]['national_peak_demand'].values[0]
            
            line_entries_95.append(df_q_95)
            line_entries_05.append(df_q_05)
            line_entries_tot_h.append(peak_weather_yr_2015)

        # Try to smooth lines
        try:
            sim_yrs_smoothed, line_entries_tot_h_smoothed = basic_plot_functions.smooth_data(sim_yrs, line_entries_tot_h, num=40000)
        except:
            sim_yrs_smoothed = sim_yrs
            line_entries_tot_h_smoothed = line_entries_tot_h


        df_q_95_scenarios = df_q_95_scenarios.append(pd.DataFrame([line_entries_95], columns=scenarios))
        df_q_05_scenarios = df_q_05_scenarios.append(pd.DataFrame([line_entries_05], columns=scenarios))
        df_total_demand_2015 = df_total_demand_2015.append(pd.DataFrame([line_entries_tot_h_smoothed], columns=scenarios))


    # ----
    # Set simulation year as index
    # ----
    df_total_demand_2015 = df_total_demand_2015.set_index([sim_yrs_smoothed])
    df_q_95_scenarios = df_q_95_scenarios.set_index([sim_yrs])
    df_q_05_scenarios = df_q_05_scenarios.set_index([sim_yrs])


    # plot lines
    for cnt, scenario in enumerate(scenarios):

        # Print total demand for specific year
        df_total_demand_2015[scenario].plot.line(color=color_list[cnt], style='-', label=": {}".format(scenario))

        # print quantiles
        #df_q_95_scenarios[scenario].plot.line(color=color_list[cnt], linestyle='--', linewidth=0.5, label="0.05")
        #df_q_05_scenarios[scenario].plot.line(color=color_list[cnt], linestyle='--', linewidth=0.5, label="0.05")
        
        # -----------------
        # Uncertainty range
        # -----------------
        plt.fill_between(
            sim_yrs, #x
            df_q_95_scenarios[scenario],  #y1
            df_q_05_scenarios[scenario],  #y2
            alpha=0.15,
            facecolor=color_list[cnt])#,
            #label=": {}".format(scenario))

    # ------------
    # Legend
    # ------------
    plt.legend(
        ncol=1,
        bbox_to_anchor=(0.5, 0., 0.5, 0.5), #bbox_to_anchor=(0.2, -0.1),
        prop={'size': 8},
        frameon=False)

    plt.ylabel("GW")
    plt.xlabel("year")

    # Tight layout
    plt.tight_layout()
    plt.margins(x=0)

    plt.show()
    prnt(".")
    # Calculate std per simulation year
    '''
    std_total_demand = list(np.std(weather_yrs_total_demand, axis=0)) # across columns calculate std
            std_peak_demand = list(np.std(weather_yrs_peak_demand, axis=0)) # across columns calculate std



            # convert to list
            df_total_demand_q_95 = df_total_demand_q_95.tolist()
            df_total_demand_q_05 = df_total_demand_q_05.tolist()
            df_peak_q_95 = df_peak_q_95.tolist()


    # -------------------
    # Base year data (2015)
    # -------------------
    # total demand
    tot_demand_twh_2015 = []
    for sim_yr, data_sim_yr in data_input[2015].items():
        gwh_2015_y = np.sum(data_sim_yr[fueltype_int])
        twh_2015_y = conversions.gwh_to_twh(gwh_2015_y)
        tot_demand_twh_2015.append(twh_2015_y)

    # peak
    df_peak_2015 = []
    for sim_yr, data_sim_yr in data_input[2015].items():
        peak_gwh_2015_y = np.max(data_sim_yr[fueltype_int])
        df_peak_2015.append(peak_gwh_2015_y)

    # ---------------
    # Smoothing lines
    # ---------------
    if len(nr_weather_yrs) > 2:
        try:
            sim_yrs_smoothed, tot_demand_twh_2015_smoothed = basic_plot_functions.smooth_data(columns, tot_demand_twh_2015, num=40000)
            period_h_smoothed, df_total_demand_q_95_smoothed = basic_plot_functions.smooth_data(list(columns), df_total_demand_q_95, num=40000)
            period_h_smoothed, df_total_demand_q_05_smoothed = basic_plot_functions.smooth_data(columns, df_total_demand_q_05, num=40000)
            period_h_smoothed, df_peak_q_95_smoothed = basic_plot_functions.smooth_data(list(columns), df_peak_q_95, num=40000)
            period_h_smoothed, df_peak_q_05_smoothed = basic_plot_functions.smooth_data(columns, df_peak_q_05, num=40000)
            period_h_smoothed, df_peak_2015_smoothed = basic_plot_functions.smooth_data(columns, df_peak_2015, num=40000)
        except:
            period_h_smoothed = columns
            df_total_demand_q_95_smoothed = df_total_demand_q_95
            df_total_demand_q_05_smoothed = df_total_demand_q_05
            df_peak_q_95_smoothed = df_peak_q_95
            df_peak_q_05_smoothed = df_peak_q_05
            tot_demand_twh_2015_smoothed = tot_demand_twh_2015
            df_peak_2015_smoothed = df_peak_2015
    else:
        try:
            period_h_smoothed, tot_demand_twh_2015_smoothed = basic_plot_functions.smooth_data(columns, tot_demand_twh_2015, num=40000)
            period_h_smoothed, df_peak_2015_smoothed = basic_plot_functions.smooth_data(columns, df_peak_2015, num=40000)
        except:
            period_h_smoothed = columns
            tot_demand_twh_2015_smoothed = tot_demand_twh_2015
            df_peak_2015_smoothed = df_peak_2015
     
    # --------------
    # Two axis figure
    # --------------
    fig, ax1 = plt.subplots(
        figsize=basic_plot_functions.cm2inch(15, 10))

    ax2 = ax1.twinx()

    # Axis label
    ax1.set_xlabel('Years')
    ax2.set_ylabel('Peak hour {} demand (GW)'.format(fueltype_str), color='black')
    ax1.set_ylabel('Total {} demand (TWh)'.format(fueltype_str), color='black')

    # Make the y-axis label, ticks and tick labels match the line color.¨
    color_axis1 = 'lightgrey'
    color_axis2 = 'blue'

    ax1.tick_params('y', colors='black')
    ax2.tick_params('y', colors='black')

    if len(nr_weather_yrs) > 2:

    '''



    pass



"""Clean original temperature data
"""
import os
import csv
from collections import defaultdict
import logging
import numpy as np
import pandas as pd

from energy_demand.basic import date_prop
from energy_demand.basic import basic_functions
from energy_demand.read_write import data_loader

def count_sequence_of_zeros(sequence):
    cnt = 0
    max_cnt_zeros = 0
    for hour_val in sequence:
        if hour_val == 0:
            cnt += 1
        else:
            if cnt > max_cnt_zeros:
                max_cnt_zeros = cnt
            cnt = 0 #reset to zero again

    return max_cnt_zeros

def nan_helper(y):
    """Helper to handle indices and logical indices of NaNs.

    Input:
        - y, 1d numpy array with possible NaNs
    Output:
        - nans, logical indices of NaNs
        - index, a function, with signature indices= index(logical_indices),
          to convert logical indices of NaNs to 'equivalent' indices
    Example:
        >>> # linear interpolation of NaNs
        >>> nans, x= nan_helper(y)
        >>> y[nans]= np.interp(x(nans), x(~nans), y[~nans])

    Taken from: https://stackoverflow.com/questions/6518811/interpolate-nan-values-in-a-numpy-array
    """
    return np.isnan(y), lambda z: z.nonzero()[0]

def run(
        path_files,
        path_out_files,
        path_weather_stations,
        crit_min_max=False
    ):
    """Iterate weather data from MIDAS and
    generate annual hourly temperatre data files
    for every weather station by interpolating missing values
    http://data.ceda.ac.uk/badc/ukmo-midas/data/WH/yearly_files/

    The data are obtained from the Centre for Environmental Data Analysis

    [Download link]http://data.ceda.ac.uk/badc/ukmo-midas/data/WH/yearly_files/ ()

    http://badc.nerc.ac.uk/artefacts/badc_datadocs/ukmo-midas/WH_Table.html (metadata)

    Weather Stations information: http://badc.nerc.ac.uk/search/midas_stations/
    http://badc.nerc.ac.uk/cgi-bin/midas_stations/search_by_name.cgi.py?name=&minyear=&maxyear=

    The values are then written to a created folder as follows:

        year__stationname.txt

    Arguments
    ---------
    path_files : str
        Path to folder with original weather data files
    path_out_files : str
        Path to folder where the cleaned data are stored

    crit_missing_values : int
        Criteria of how many missing values there mus bet until
        a weather station is discarded
    crit_nr_of_zeros : int
        Criteria of how many zeros there must be per year until
        weather station is discarde
    nr_daily_zeros : int
        How many zero values there can be maxmum in a day in the
        year until the station is discarded

    Note
    -----
        - In case of a leap year the 29 of February is ignored
    """
    logging.info("... starting to clean original weather files")

    # Load coordinates of weather stations
    weather_stations = data_loader.read_weather_stations_raw(path_weather_stations)

    # Stations which are outisde of the uk and are ignored
    stations_outside_UK = [
        1605, # St. Helena
        1585, # Gibraltar
        1609] # Falkland Islands

    # Create out folder to store station specific csv
    if os.path.isdir(path_out_files):
        basic_functions.delete_folder(path_out_files)
        os.mkdir(path_out_files)
    else:
        os.mkdir(path_out_files)

    # Placeholder value for missing entry
    placeholder_value = np.nan

    # Read all files
    all_annual_raw_files = os.listdir(path_files)

    # Sort according to year
    all_annual_raw_files.sort()

    # Annual temperature file
    for file_name in all_annual_raw_files:
        logging.info("... reading in file: %s", file_name)

        path_to_csv = os.path.join(path_files, file_name)
        temp_stations = {}
        temp_stations_min_max = defaultdict(dict)

        with open(path_to_csv, 'r') as csvfile:
            read_lines = csv.reader(csvfile, delimiter=',')

            for row in read_lines:
                date_measurement = row[0].split(" ")
                year = int(date_measurement[0].split("-")[0])
                month = int(date_measurement[0].split("-")[1])
                day = int(date_measurement[0].split("-")[2])
                hour = int(date_measurement[1][:2])

                yearday = date_prop.date_to_yearday(year, month, day)

                if date_prop.is_leap_year(year):
                    yearday = yearday - 1 # Substract because 29. of Feb is later ignored
                else:
                    pass

                year_hour = (yearday * 24) + hour

                # Weather station id
                station_id = int(row[5])

                # If station is outside uk or leap year day
                if (station_id in stations_outside_UK) or (month == 2 and day == 29):
                    pass
                else:
                    # Air temperature in Degrees Celcius
                    if row[35] == ' ' or row[35] == '': # If no data point
                        air_temp = placeholder_value
                    else:
                        air_temp = float(row[35])

                    # Add weather station if not already added to dict
                    if station_id not in temp_stations:
                        if crit_min_max:
                            temp_stations_min_max[station_id]['t_min'] = np.zeros((365), dtype="float")
                            temp_stations_min_max[station_id]['t_max'] = np.zeros((365), dtype="float")
                            temp_stations[station_id] = []
                    else:
                        pass

                    if air_temp is not placeholder_value:

                        if yearday not in temp_stations[station_id]:
                            temp_stations_min_max[station_id]['t_min'][yearday] = air_temp
                            temp_stations_min_max[station_id]['t_max'][yearday] = air_temp
                            temp_stations[station_id].append(yearday)

                        if crit_min_max:
                            # Update min and max daily temperature
                            if air_temp < temp_stations_min_max[station_id]['t_min'][yearday]:
                                temp_stations_min_max[station_id]['t_min'][yearday] = air_temp
                            if air_temp > temp_stations_min_max[station_id]['t_max'][yearday]:
                                temp_stations_min_max[station_id]['t_max'][yearday] = air_temp

        # ------------------------
        # Delete weather stations with missing daily data inputs
        # ------------------------
        stations_to_delete = []
        for station_id in temp_stations_min_max.keys():

            nans, x = nan_helper(temp_stations_min_max[station_id]['t_min'])
            nr_of_nans_t_min = list(nans).count(True)

            nans, x = nan_helper(temp_stations_min_max[station_id]['t_max'])
            nr_of_nans_t_max = list(nans).count(True)

            if nr_of_nans_t_min > 0 or nr_of_nans_t_max > 0:
                print("Station '{}' contains {} {} .nan values and is deleted".format(station_id, nr_of_nans_t_min, nr_of_nans_t_max))
                stations_to_delete.append(station_id)
        
        print("Number of stations to delete: {}".format(len(stations_to_delete)))
        for i in stations_to_delete:
            del temp_stations_min_max[i]

        # --------------------
        # Write out files
        # --------------------
        path_out_stations = os.path.join(path_out_files, '{}_stations.csv'.format(str(year)))
        path_out_t_min = os.path.join(path_out_files, "{}_t_min.npy".format(str(year)))
        path_out_t_max = os.path.join(path_out_files, "{}_t_max.npy".format(str(year)))

        # Check if weather station is defined
        stations_to_delete = []
        for name in temp_stations_min_max.keys():
            try:
               _ = weather_stations[name]['latitude']
               _ = weather_stations[name]['longitude']
            except KeyError:
                print("... no coordinates are available for weather station '{}'".format(name))
                stations_to_delete.append(name)

        print("... number of stations to delete: {}".format(len(stations_to_delete)))
        for name in stations_to_delete:
            del temp_stations_min_max[name]

        stations = list(temp_stations_min_max.keys())
        out_list = []
        for station_name in stations:
            out_list.append([station_name, weather_stations[station_name]['latitude'], weather_stations[station_name]['longitude']])

        # Write
        df = pd.DataFrame(np.array(out_list), columns=['station_id', 'latitude', 'longitude'])
        df.to_csv(path_out_stations, index=False)

        stations_t_min = list(i['t_min'] for i in temp_stations_min_max.values())
        stations_t_max = list(i['t_max'] for i in temp_stations_min_max.values())
        stations_t_min = np.array(stations_t_min)
        stations_t_max = np.array(stations_t_max)

        np.save(path_out_t_min, stations_t_min)
        np.save(path_out_t_max, stations_t_max)

    logging.info("... finished cleaning weather data")

run(
    path_files="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/_meteo_data_2015",
    path_out_files="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/_complete_meteo_data_all_yrs_cleaned_min_max",
    path_weather_stations="//linux-filestore.ouce.ox.ac.uk/mistral/nismod/data/energy_demand/H-Met_office_weather_data/cleaned_weather_stations.csv",
    crit_min_max=True)
